from art.attacks.evasion import FastGradientMethod
from ..metrics import true_min_pert
import numpy as np


class LpAttack:
    def __init__(self, model, attack, use_model_preds=False, **attkwargs):
        """
        Makes a generic L^p attacker instance.

        :param model: TrojClassifier
        :param attack: an undeclared instance of an ART evasion attack i.e art.evasion.pgd
        :param attkwargs: keyword arguments for the attack
        """
        self.model = model
        attack_dict = {}
        att_name = "attack"
        attack_dict[att_name] = attack(model, **attkwargs)
        self.attack_dict = attack_dict
        self.use_model_preds = use_model_preds

    def generate(self, x, y):
        """
        :param x: inputs
        :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
        the first dimension because of art bug.
        :return: adversarial examples with minimal perturbation, adversarial losses, adversarial predictions
        """
        generated_examples = []
        model_adv_losses = []
        model_adv_preds = []
        # For each attack method in the attack dictionary, generate adversarial examples, then compute the loss
        # and the class predictions.
        for attacker in list(self.attack_dict.values()):
            adv_x = attacker.generate(x, y)
            if len(y.shape) == 1:
                adv_losses, adv_preds = self.model.ComputeLoss(adv_x, y)
                if self.use_model_preds == True:
                    true_losses, true_preds = self.model.ComputeLoss(x, y)
            else:
                adv_losses, adv_preds = self.model.ComputeLoss(
                    adv_x, np.squeeze(y, axis=1)
                )
                if self.use_model_preds == True:
                    true_losses, true_preds = self.model.ComputeLoss(
                        x, np.squeeze(y, axis=1)
                    )
            model_adv_losses.append(adv_losses)
            generated_examples.append(adv_x)
            model_adv_preds.append(adv_preds)
        # reshape arrays so that they each have shape [batch_size, num_attacks, *]
        generated_examples = np.stack(generated_examples)
        generated_examples = np.swapaxes(generated_examples, 0, 1)

        model_adv_losses = np.stack(model_adv_losses)
        model_adv_losses = np.swapaxes(model_adv_losses, 0, 1)

        model_adv_preds = np.stack(model_adv_preds)
        model_adv_preds = np.swapaxes(model_adv_preds, 0, 1)

        # Whether or not to compute the perturbations quickly, or in such a way as to minimize
        # the perturbation that flips the label.
        if self.use_model_preds == True:
            labels = true_preds
        else:
            labels = y
        output, preds, losses = true_min_pert(
            labels, model_adv_preds, x, generated_examples, model_adv_losses
        )

        return output, preds, losses


class TrojEpsAttack:
    def __init__(
        self,
        model,
        eps_steps=0.01,
        max_eps=0.5,
        batch_size=128,
        norm=np.inf,
        use_model_preds=False,
    ):
        """
        This is the basic evaluation attack, and is equivalent to
         LpAttack(model, art.evasion.fgm, attkwargs = {'minimal'=True, ...}). This is meant to provide a simple evaluation
         of a classifier against the simplest whitebox attack algorithm.


        :param model: Troj/ART classifier instance.
        :param eps_steps: Size of steps in FGSM
        :param max_eps: Largest allowable perturbation
        :param batch_size: Batch size for each attack.
        :param norm: The attack norm for FGSM.
        :param use_model_preds: if minimum is true, whether or not to use the true labels or the model preds for
        evauluating the minimum perturbation
        """
        self.model = model
        self.max_eps = max_eps
        self.eps_steps = eps_steps
        self.batch_size = batch_size
        self.norm = norm
        self.use_model_preds = use_model_preds

        # store instantiate attack instances in dictionary
        attack_dict = {}
        fgm_name = "fgsm_{}".format(0)
        attack_dict[fgm_name] = FastGradientMethod(
            self.model,
            norm=self.norm,
            eps=self.max_eps,
            eps_step=self.eps_steps,
            batch_size=self.batch_size,
            minimal=True,
        )

        self.attack_dict = attack_dict

    def generate(self, x, y):
        """
        :param x: inputs
        :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
        the first dimension because of art bug.
        :return: adversarial examples with minimal perturbation, adversarial losses, adversarial predictions
        """
        generated_examples = []
        model_adv_losses = []
        model_adv_preds = []
        # For each attack method in the attack dictionary, generate adversarial examples, then compute the loss
        # and the class predictions.
        for attacker in list(self.attack_dict.values()):
            adv_x = attacker.generate(x, y)
            if len(y.shape) == 1:
                adv_losses, adv_preds = self.model.ComputeLoss(adv_x, y)
                if self.use_model_preds == True:
                    true_losses, true_preds = self.model.ComputeLoss(x, y)
            else:
                adv_losses, adv_preds = self.model.ComputeLoss(
                    adv_x, np.squeeze(y, axis=1)
                )
                if self.use_model_preds == True:
                    true_losses, true_preds = self.model.ComputeLoss(
                        x, np.squeeze(y, axis=1)
                    )
            model_adv_losses.append(adv_losses)
            generated_examples.append(adv_x)
            model_adv_preds.append(adv_preds)
        # reshape arrays so that they each have shape [batch_size, num_attacks, *]
        generated_examples = np.stack(generated_examples)
        generated_examples = np.swapaxes(generated_examples, 0, 1)

        model_adv_losses = np.stack(model_adv_losses)
        model_adv_losses = np.swapaxes(model_adv_losses, 0, 1)

        model_adv_preds = np.stack(model_adv_preds)
        model_adv_preds = np.swapaxes(model_adv_preds, 0, 1)

        # Whether or not to compute the perturbations quickly, or in such a way as to minimize
        # the perturbation that flips the label.
        if self.use_model_preds == True:
            labels = true_preds
        else:
            labels = y
        output, preds, losses = true_min_pert(
            labels, model_adv_preds, x, generated_examples, model_adv_losses
        )

        return output, preds, losses



import re
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from requests import Session
from typing import Any, Dict, List
import os
import requests
import json
import zipfile

retry_strategy = Retry(
    total=2,
    backoff_factor=1,
    status_forcelist=[429, 500, 502, 503, 504],
    method_whitelist=["HEAD", "GET", "PUT", "POST", "DELETE", "OPTIONS", "TRACE"],
)

retry_adapter = HTTPAdapter(max_retries=retry_strategy)
requests_retry = Session()
requests_retry.mount("https://", retry_adapter)
requests_retry.mount("http://", retry_adapter)
tempdir_ttl_days = 1


def assert_valid_name(name: str):
    is_valid = re.match(r"^[A-Za-z0-9_]+$", name)
    if not is_valid:
        raise Exception(
            f"'{name}' must only contain alphanumeric and underscore characters."
        )


def raise_resp_exception_error(resp):
    if not resp.ok:
        message = None
        try:
            r_body = resp.json()
            message = r_body.get("message") or r_body.get("msg")
        except:
            # If we failed for whatever reason (parsing body, etc.)
            # Just return the code
            if resp.status_code == 500:
                raise Exception(
                    f"HTTP Error received: {resp.reason}: {str(resp.status_code)}"
                )
            else:
                raise Exception(
                    f"HTTP Error received: {resp.reason}: {str(resp.status_code)} | {resp.json()['detail']}"
                )
        if message:
            raise Exception(f"Error: {message}")
        else:
            if resp.status_code == 500:
                raise Exception(
                    f"HTTP Error received: {resp.reason}: {str(resp.status_code)}"
                )
            else:
                raise Exception(
                    f"HTTP Error received: {resp.reason}: {str(resp.status_code)} | {resp.json()['detail']}"
                )


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def _upload_local_files(
    file_names: List[str],
    get_upload_path: str,
    headers: Dict[str, Any],
    upload_prefix: str,
    upload_suffix: str,
    delete_after_upload: bool = True,
):
    """
    This uploads a set of files in batches with a reader.

    Args:
        file_names (str): The local file_names (files to be uploaded)
        get_upload_path (str): The URL that generates upload URLs
        headers (Dict[str, Any]): Headers for the get_upload_path request
        upload_prefix (str): Prefix for the filepath (once uploaded)
        upload_suffix (str): Suffix for the filepath (once uploaded)
        delete_after_upload (bool): Whether to delete the file after upload

    Return:
        A list of download URLs for the uploaded files
    """
    xml_api_headers = {
        "content-type": "application/octet-stream",
    }

    download_urls = []
    if len(file_names) == 0:
        return download_urls

    url_chunks = chunks(file_names, 100)

    param_batch_list = []
    for i, chunk in enumerate(url_chunks, start=1):
        for count, file_name in enumerate(chunk, start=1):

            upload_filename, ext = os.path.splitext(os.path.basename(file_name))

            param_batch_list.append(
                {"file_name": upload_filename, "file_extension": ext}
            )

            if count == len(chunk):

                data = {"files": param_batch_list}

                upload_urls_resp = requests_retry.post(
                    get_upload_path,
                    headers=headers,
                    data=json.dumps(data).encode("utf-8"),
                )
                raise_resp_exception_error(upload_urls_resp)
                batch_of_urls = upload_urls_resp.json()
                param_batch_list = []
        # batch_of_urls comes back in the same expected order as the chunk, so we zip and enumerate simultaneously.
        for count, (file_name, urlobj) in enumerate(zip(chunk, batch_of_urls), start=1):
            with open(file_name, "rb") as f:
                files = {"file": (file_name, f)}
                http_response = requests.post(
                    urlobj["url"], data=urlobj["fields"], files=files
                )

    return True


def _upload_local_zip(
    zip_filepath: str,
    get_upload_path: str,
    complete_upload_path: str,
    headers: Dict[str, Any],
    upload_prefix: str,
    upload_suffix: str,
    delete_after_upload: bool = True,
):
    """This uploads a set of files in batches with a reader.

    Args:
        file_names (str): The local file_names (files to be uploaded)
        get_upload_path (str): The URL that generates upload URLs
        headers (Dict[str, Any]): Headers for the get_upload_path request
        upload_prefix (str): Prefix for the filepath (once uploaded)
        upload_suffix (str): Suffix for the filepath (once uploaded)
        delete_after_upload (bool): Whether to delete the file after upload

    Return:
        A list of download URLs for the uploaded files
    """
    xml_api_headers = {
        "content-type": "application/octet-stream",
    }

    download_urls = []

    upload_filename, ext = os.path.splitext(os.path.basename(zip_filepath))

    max_size = 1 * 1024 ** 3  # 5GB is max, we use 1GB
    zip_size = os.path.getsize(zip_filepath)
    parts_no = zip_size // max_size + 1
    num_files = len(zipfile.ZipFile(zip_filepath).namelist())
    payload = {"parts": parts_no, "num_files": num_files}

    upload_url_parts_resp = requests_retry.get(
        get_upload_path, headers=headers, params=payload
    )

    raise_resp_exception_error(upload_url_parts_resp)
    upload_id = upload_url_parts_resp.json()[0]
    batch_of_urls = upload_url_parts_resp.json()[1]

    parts = []
    with open(zip_filepath, "rb") as fin:
        for num, url in enumerate(batch_of_urls):
            part = num + 1
            file_data = fin.read(max_size)

            res = requests.put(url, data=file_data)
            # Should this raise an Exception instead of just returning?
            if res.status_code != 200:
                return
            etag = res.headers["ETag"]
            parts.append({"ETag": etag, "PartNumber": part})

    data = {"parts": parts}

    comp_resp = requests_retry.post(
        complete_upload_path + "/" + upload_id,
        headers=headers,
        data=json.dumps(data).encode("utf-8"),
    )

    return {"status_code": comp_resp.status_code, "data": comp_resp.json()}


def validate_dataservices_args(**kwargs):
    less_than_equal_one = ["momentum", "adam_eps", "weight_decay", "factor"]

    if kwargs["early_stop_patience"] > kwargs["num_epochs"]:
        raise ValueError(
            f"'early_stop_patience' is {kwargs['early_stop_patience']} and must be equal to or less than 'num_epochs' which is {kwargs['num_epochs']}."
        )

    for arg_name, arg_value in kwargs.items():
        # Checking int/float values
        if type(arg_value) is int or type(arg_value) is float:
            if arg_name in less_than_equal_one:
                if arg_value < 0 or arg_value > 1:
                    raise ValueError(
                        f"{arg_name} is {arg_value} and must be between 0-1."
                    )

            if arg_value < 0:
                raise ValueError(
                    f"{arg_name} is {arg_value} and must be equal to or greater than 0."
                )
        # Checking tuple values
        if type(arg_value) is tuple:
            for num in arg_value:
                if num < 0:
                    raise ValueError(
                        f"{arg_name} is {arg_value} and must have values equal to or greater than 0."
                    )



import torch
import torch.nn as nn


class PytorchObjectDetector(nn.Module):
    def __init__(self, model):
        """

        :param model: Pytorch Object Detection model
        """
        super(PytorchObjectDetector, self).__init__()
        self.model = model

    def compute_loss(self, samples, labels, grad=False, reduce=True):
        """
        :param: samples: Samples to make predictions on as list of tensors.
        :param: labels: labels as list of dictionaries
        :param: grad: If true, the gradient is computed.
        :param: reduce: If true, the losses are reduced via summation and a single scalar is returned, otherwise a dictionary is returned.
        """
        self.model.train()
        if grad == False:
            with torch.no_grad():
                losses = self.model(samples, labels)
        else:
            losses = self.model(samples, labels)
        if reduce:
            losses = sum(loss for loss in losses.values())
        return losses

    def predict(self, sample):
        self.model.eval()
        preds = self.model(sample)
        return preds



from ..attacks.ClassifierAttacks import LpAttack, TrojEpsAttack
from ..metrics.metric_utils import compute_Lp_distance
import numpy as np


class GenericRobustnessEvaluator:
    def __init__(
        self, classifier, attack, attack_name, attkwargs, use_model_preds=False
    ):
        """
        The simple robustness evaluator class. Should be used to form a baseline.

        :param classifier: TrojClassifier/ART Classifier instance
        :param attack: undecalred ART evasion class
        :param attack_name: Name for logging
        :param **attkwargs: keyword arguments for attack.
        """
        self.atk_meta = attkwargs
        self.classifier = classifier
        self.attacker = LpAttack(self.classifier, attack, use_model_preds, **attkwargs)
        self.atk_meta["name"] = attack_name

    def attack(self, data, target, index, device=None):
        """
        Runs the attack.

        :param data: Input data
        :param target: Target labels
        :param index: Index of samples in dataframe
        :param device: If using Pytorch, one can specify the device.
        :return: A dictionary containing the minimum perturbation, the loss, adversarial loss, prediction, and adversarial
        prediction for each sample, along with a list of indices for the logging function.
        """

        # send data and target to cpu, convert to numpy
        data = np.ascontiguousarray(data.astype(np.float32))
        test_loss, preds = self.classifier.ComputeLoss(data, target)
        preds = np.argmax(preds, axis=1)
        adv_x, adv_preds, adv_loss = self.attacker.generate(data, target)
        # adv_loss, adv_preds = self.classifier.ComputeLoss(adv_x, target)
        perturbation = compute_Lp_distance(data, adv_x)
        adv_pred = np.argmax(adv_preds, axis=1)
        # generate the adversarial image using the data numpy array and label numpy array
        out_dict = {
            "Linf_perts": perturbation,
            "Loss": test_loss,
            "Adversarial_Loss": adv_loss,
            "prediction": preds,
            "Adversarial_prediction": adv_pred,
        }
        return (out_dict, index)


class BasicRobustnessEvaluator:
    def __init__(
        self,
        classifier,
        eps_steps=0.01,
        batch_size=128,
        norm=np.inf,
        use_model_preds=False,
    ):
        """
        The simple robustness evaluator class. Should be used to form a baseline.

        :param classifier: TrojClassifier/ART Classifier instance
        :param eps_steps: Epsilon step size for the attack. Smaller values are more precise but slower.
        :param batch_size: Number of images in a batch.
        :param norm: The p norm (can be int or np.inf)
        :param use_model_preds: Whether or not to use true labels or model predictions
        """
        self.atk_meta = {
            "attack_name": "fgsm",
            "eps_steps": str(eps_steps),
            "batch_size": str(batch_size),
            "norm": str(norm),
            "use_model_preds": str(use_model_preds),
        }
        self.classifier = classifier
        self.attacker = TrojEpsAttack(
            self.classifier,
            eps_steps=eps_steps,
            batch_size=batch_size,
            norm=norm,
            use_model_preds=use_model_preds,
        )

    def attack(self, data, target, index, device=None):
        """
        Runs the attack.

        :param data: Input data
        :param target: Target labels
        :param index: Index of samples in dataframe
        :param device: If using Pytorch, one can specify the device.
        :return: A dictionary containing the minimum perturbation, the loss, adversarial loss, prediction, and adversarial
        prediction for each sample, along with a list of indices for the logging function.
        """

        # send data and target to cpu, convert to numpy
        data = np.ascontiguousarray(data.astype(np.float32))
        test_loss, preds = self.classifier.ComputeLoss(data, target)
        preds = np.argmax(preds, axis=1)
        adv_x, adv_preds, adv_loss = self.attacker.generate(data, target)
        # adv_loss, adv_preds = self.classifier.ComputeLoss(adv_x, target)
        perturbation = compute_Lp_distance(data, adv_x)
        adv_pred = np.argmax(adv_preds, axis=1)
        # generate the adversarial image using the data numpy array and label numpy array
        out_dict = {
            "Linf_perts": perturbation,
            "Loss": test_loss,
            "Adversarial_Loss": adv_loss,
            "prediction": preds,
            "Adversarial_prediction": adv_pred,
        }
        return (out_dict, index)



import numpy as np
from PIL import Image
import os
import random
from .Corruptions import *
import json
import copy


corruption_tuple = (gaussian_noise, shot_noise, impulse_noise, defocus_blur,
                    glass_blur, motion_blur, zoom_blur, snow, frost, fog,
                    brightness, contrast, elastic_transform, pixelate,
                    jpeg_compression, speckle_noise, gaussian_blur, spatter,
                    saturate)

corruption_dict = {corr_func.__name__: corr_func for corr_func in
                   corruption_tuple}


def corrupt(image, severity=1, corruption_name=None, corruption_number=-1):
    """This function returns a corrupted version of the given image.

    Args:
        image (numpy.ndarray):      image to corrupt; a numpy array in [0, 255], expected datatype is np.uint8
                                    expected shape is either (height x width x channels) or (height x width);
                                    width and height must be at least 32 pixels;
                                    channels must be 1 or 3;
        severity (int):             strength with which to corrupt the image; an integer in [1, 5]
        corruption_name (str):      specifies which corruption function to call, must be one of
                                        'gaussian_noise', 'shot_noise', 'impulse_noise', 'defocus_blur',
                                        'glass_blur', 'motion_blur', 'zoom_blur', 'snow', 'frost', 'fog',
                                        'brightness', 'contrast', 'elastic_transform', 'pixelate', 'jpeg_compression',
                                        'speckle_noise', 'gaussian_blur', 'spatter', 'saturate';
                                    the last four are validation corruptions
        corruption_number (int):    the position of the corruption_name in the above list; an integer in [0, 18];
                                        useful for easy looping; 15, 16, 17, 18 are validation corruption numbers
    Returns:
        numpy.ndarray:              the image corrupted by a corruption function at the given severity; same shape as input
    """

    if not isinstance(image, np.ndarray):
        raise AttributeError('Expecting type(image) to be numpy.ndarray')
    if not (image.dtype.type is np.uint8):
        raise AttributeError('Expecting image.dtype.type to be numpy.uint8')

    if not (image.ndim in [2, 3]):
        raise AttributeError('Expecting image.shape to be either (height x width) or (height x width x channels)')
    if image.ndim == 2:
        image = np.stack((image,) * 3, axis=-1)

    height, width, channels = image.shape

    if (height < 32 or width < 32):
        raise AttributeError('Image width and height must be at least 32 pixels')

    if not (channels in [1, 3]):
        raise AttributeError('Expecting image to have either 1 or 3 channels (last dimension)')

    if channels == 1:
        image = np.stack((np.squeeze(image),) * 3, axis=-1)

    if not severity in [1, 2, 3, 4, 5]:
        raise AttributeError('Severity must be an integer in [1, 5]')

    if not (corruption_name is None):
        image_corrupted = corruption_dict[corruption_name](Image.fromarray(image),
                                                           severity)
    elif corruption_number != -1:
        image_corrupted = corruption_tuple[corruption_number](Image.fromarray(image),
                                                              severity)
    else:
        raise ValueError("Either corruption_name or corruption_number must be passed")

    return np.uint8(image_corrupted)


def get_corruption_names(subset='common'):
    if subset == 'common':
        return [f.__name__ for f in corruption_tuple[:15]]
    elif subset == 'validation':
        return [f.__name__ for f in corruption_tuple[15:]]
    elif subset == 'all':
        return [f.__name__ for f in corruption_tuple]
    elif subset == 'noise':
        return [f.__name__ for f in corruption_tuple[0:3]]
    elif subset == 'blur':
        return [f.__name__ for f in corruption_tuple[3:7]]
    elif subset == 'weather':
        return [f.__name__ for f in corruption_tuple[7:11]]
    elif subset == 'digital':
        return [f.__name__ for f in corruption_tuple[11:15]]
    else:
        raise ValueError("subset must be one of ['common', 'validation', 'all']")





def im_loc_splitter_classification(location):
    '''
    Retrieves the class of the image and the filename for classification datasets
    :param location: Location of image file.
    :returns: The class of the image and the filename
    '''
    folderwise_split = location.split('/')
    img_file = folderwise_split[-1]
    img_class = folderwise_split[-2]
    return img_file, img_class

def rename_img(img_file, corruption_name, severity):
    '''
    :param img_file: name of the image file
    :param corruption_name: corruption name
    :param severity: severity of the corruption
    :returns: new file name for corrupt image in format {original name}_cor_{corruption name}_sev_{severity of corruption}.extension
    '''
    split_file_extension = img_file.split('.')
    file_extension = '.{}'.format(split_file_extension[-1])
    img_name_extension = '_cor_{}_sev_{}'.format(corruption_name, severity)
    corrupt_img_name = split_file_extension[0] + img_name_extension + file_extension
    return corrupt_img_name

def corrupt_and_save_classification(original_location, new_root, corruption, severity, image_array=None):
    '''
    :param original_location: Original image location.
    :param new_root: The new root folder.
    :param corruption: The corruption to apply.
    :param severity: The severity of the corruption.
    :param image_array: A preloaded image array, if none, the image is loaded from the original_location parameter
    :returns: saves corrupt image at the location {new_root}/{class}/{corrupt image name}

    '''
    im_file, im_class = im_loc_splitter_classification(original_location)
    corrupt_im_name = rename_img(im_file, corruption, severity)
    save_loc = os.path.join(new_root, im_class)
    if os.path.exists(new_root)==False:
        os.makedirs(new_root)
    if os.path.exists(os.path.join(new_root, im_class))==False:
        os.makedirs(save_loc)
    if image_array is None:
        image_array=np.asarray(Image.open(original_location))
    original_image = image_array
    corrupted = corrupt(original_image, corruption_name=corruption, severity=severity)
    corrupted_image = Image.fromarray(corrupted)
    corrupted_image.save(os.path.join(save_loc, corrupt_im_name))


def corrupt_dset_random(original_folder, new_folder, corruptions, severities,  include_original=False, verbose=True):
    '''
    :param original_folder: The original split folder (test folder).
    :param new_folder: New folder location to save corrupt images to.
    :param corruptions: A list of corruptions.
    :param severities: A list of severity levels.
    :param include_original: Whether or not to include the original image in the new test set.
    :param verbose: Whether or not to print generation progress.
    :returns: Creates a new dataset at new_folder containing images which are corrupted according to a random sampling
    of the corruptions and severity
    '''
    classes = os.listdir(original_folder)
    num_classes = len(classes)
    class_counter = 0
    if os.path.exists(new_folder)==False:
        os.makedirs(new_folder)
    for class_name in classes:
        class_counter += 1
        im_counter = 0
        class_folder = os.path.join(original_folder, class_name)
        im_files = os.listdir(class_folder)
        num_imgs = len(im_files)
        new_class_folder = os.path.join(new_folder, class_name)
        if os.path.exists(new_class_folder) == False:
            os.makedirs(new_class_folder)
        for img_file in im_files:
            im_counter += 1
            im_loc = os.path.join(class_folder, img_file)
            image = Image.open(im_loc)
            if include_original==True:
                uncorrupt_name= rename_img(img_file, 'none', 0)
                image.save(os.path.join(new_class_folder, uncorrupt_name))
            image_arr = np.asarray(image)
            corrupt_name = random.choice(corruptions)
            severity = random.choice(severities)
            corrupt_and_save_classification(im_loc, new_folder, corrupt_name, severity, image_arr)
            if verbose==True:
                progress_str = 'Class {}/{}: Image {}/{}'.format(class_counter, num_classes, im_counter, num_imgs)
                print(progress_str)


def corrupt_dset_all(original_folder, new_folder, corruptions, severities,  include_original=False, verbose=True):
    '''
    :param original_folder: The original split folder (test folder).
    :param new_folder: New folder location to save corrupt images to.
    :param corruptions: A list of corruptions.
    :param severities: A list of severity levels.
    :param include_original: Whether or not to include the original image in the new test set.
    :param verbose: Whether or not to print generation progress.
    :returns: Creates a new dataset at new_folder where every image is corrupted by each corruption and each severity and saved.
    '''
    classes = os.listdir(original_folder)
    num_classes = len(classes)
    class_counter = 0
    if os.path.exists(new_folder)==False:
        os.makedirs(new_folder)
    for class_name in classes:
        class_counter += 1
        im_counter = 0
        class_folder = os.path.join(original_folder, class_name)
        im_files = os.listdir(class_folder)
        num_imgs = len(im_files)
        new_class_folder = os.path.join(new_folder, class_name)
        if os.path.exists(new_class_folder) == False:
            os.makedirs(new_class_folder)
        for img_file in im_files:
            im_counter += 1
            im_loc = os.path.join(class_folder, img_file)
            image = Image.open(im_loc)
            if include_original==True:
                uncorrupt_name= rename_img(img_file, 'none', 0)
                image.save(os.path.join(new_class_folder, uncorrupt_name))
            image_arr = np.asarray(image)
            for corrupt_name in corruptions:
                for severity in severities:
                    corrupt_and_save_classification(im_loc, new_folder, corrupt_name, severity, image_arr)
            if verbose==True:
                progress_str = 'Class {}/{}: Image {}/{}'.format(class_counter, num_classes, im_counter, num_imgs)
                print(progress_str)



def MakeCorruptDsetClassification(original_folder, new_folder, corruptions, severities, mode='random',  include_original=False, verbose=True):
    '''
    :param original_folder: The original split folder (test folder).
    :param new_folder: New folder location to save corrupt images to.
    :param corruptions: A list of corruptions.
    :param severities: A list of severity levels.
    :param mode: One of either 'random' or 'all'. If random, each image is corrupted by randomly sampling a corruption and severity level.
    If all, every corruption and every severity is applied to each image one at a time and saved.
    :param include_original: Whether or not to include the original image in the new test set.
    :param verbose: Whether or not to print generation progress.
    :returns: Creates a new dataset at new_folder containing corrupt images.
    '''
    if mode.lower()=='random':
        corrupt_dset_random(original_folder, new_folder, corruptions, severities, include_original, verbose)
    if mode.lower()=='all':
        corrupt_dset_all(original_folder, new_folder, corruptions, severities, include_original, verbose)



def change_annotation_imfile(annot_dict, corruption_name, severity):
    annot_dict_copy = copy.deepcopy(annot_dict)
    img_file = annot_dict['file_name']
    corrupt_name = rename_img(img_file, corruption_name, severity)
    annot_dict_copy['file_name'] = corrupt_name
    return annot_dict_copy

def corrupt_and_save_OD(annot_dict, old_root, new_root, corruption, severity, image_array=None):
    '''
    :param annot_dict: Annotation dictionary for the example
    :param old_root: root folder for the old image
    :param new_root: The new root folder.
    :param corruption: The corruption to apply.
    :param severity: The severity of the corruption.
    :param image_array: A preloaded image array, if none, the image is loaded from the original_location parameter
    :returns: saves corrupt image at the location {new_root}/{corrupt image name}, and returns the new annotation dictionary
    '''
    original_location = os.path.join(old_root, annot_dict['file_name'])
    new_annotation_dict = change_annotation_imfile(annot_dict, corruption, severity)
    save_loc = os.path.join(new_root, new_annotation_dict['file_name'])
    if os.path.exists(new_root)==False:
        os.makedirs(new_root)
    if image_array is None:
        image_array=np.asarray(Image.open(original_location))
    original_image = image_array
    corrupted = corrupt(original_image, corruption_name=corruption, severity=severity)
    corrupted_image = Image.fromarray(corrupted)
    corrupted_image.save(save_loc)
    return new_annotation_dict

def corrupt_od_dset_random(original_folder, new_folder, annot_save_loc, annotations, corruptions, severities, include_original=False, verbose=True):
    '''
    :param original_folder: Original folder location.
    :param new_folder: Folder location for new images
    :param annot_save_loc: File path for new annotations, saved in Troj format.
    :param annotations: Annotations dictionary in Troj format.
    :param corruptions: List of corruptions.
    :param severities: List of severities
    :param include_original: Whether or not to include the original image in the new dataset.
    :param verbose: Whether or not to print the progress of the corruptions.
    '''
    corruptions_annot_dict = {}
    counter = 0
    if os.path.exists(new_folder)==False:
        os.makedirs(new_folder)
    for im_key in list(annotations.keys()):
        im_annots = annotations[im_key]
        corrupt_name = random.choice(corruptions)
        severity = random.choice(severities)

        original_im_loc = os.path.join(original_folder, im_annots['file_name'])
        image = Image.open(original_im_loc)
        if include_original:
            corruptions_annot_dict[str(im_key)] = im_annots
            image.save(os.path.join(new_folder, im_annots['file_name']))
        image_arr = np.asarray(image)
        new_annots = change_annotation_imfile(im_annots, corrupt_name, severity)
        new_key = str(im_key) + '_cor_{}_sev_{}'.format(corrupt_name, severity)
        corruptions_annot_dict[new_key] = new_annots
        corrupted = corrupt(image_arr, corruption_name=corrupt_name, severity=severity)
        corrupted_image = Image.fromarray(corrupted)
        corrupted_image.save(os.path.join(new_folder, new_annots['file_name']))
        if verbose:
            print('{}/{} Images corrupted and saved.'.format(counter, len(annotations)))
            counter += 1
    with open(annot_save_loc, 'w') as fp:
        json.dump(corruptions_annot_dict, fp)
    return corruptions_annot_dict

def corrupt_od_dset_all(original_folder, new_folder, annot_save_loc, annotations, corruptions, severities, include_original=False, verbose=True):
    '''
    :param original_folder: Original folder location.
    :param new_folder: Folder location for new images
    :param annot_save_loc: File path for new annotations, saved in Troj format.
    :param annotations: Annotations dictionary in Troj format.
    :param corruptions: List of corruptions.
    :param severities: List of severities
    :param include_original: Whether or not to include the original image in the new dataset.
    :param verbose: Whether or not to print the progress of the corruptions.
    '''
    corruptions_annot_dict = {}
    counter = 0
    if os.path.exists(new_folder)==False:
        os.makedirs(new_folder)
    for im_key in list(annotations.keys()):
        im_annots = annotations[im_key]

        original_im_loc = os.path.join(original_folder, im_annots['file_name'])
        image = Image.open(original_im_loc)
        if include_original:
            corruptions_annot_dict[str(im_key)] = im_annots
            image.save(os.path.join(new_folder, im_annots['file_name']))
        image_arr = np.asarray(image)
        for corrupt_name in corruptions:
            for severity in severities:
                new_annots = change_annotation_imfile(im_annots, corrupt_name, severity)
                new_key = str(im_key) + '_cor_{}_sev_{}'.format(corrupt_name, severity)
                corruptions_annot_dict[new_key] = new_annots
                corrupted = corrupt(image_arr, corruption_name=corrupt_name, severity=severity)
                corrupted_image = Image.fromarray(corrupted)
                corrupted_image.save(os.path.join(new_folder, new_annots['file_name']))
        if verbose:
            print('{}/{} Images corrupted and saved.'.format(counter, len(annotations)))
            counter += 1
    with open(annot_save_loc, 'w') as fp:
        json.dump(corruptions_annot_dict, fp)
    return corruptions_annot_dict

def MakeCorruptDsetOD(original_folder, new_folder, annot_save_loc, annotations, corruptions, severities, mode='random',  include_original=False, verbose=True):
    '''
    :param original_folder: The original split folder (test folder).
    :param new_folder: New folder location to save corrupt images to.
    :param annot_save_loc: File path for new annotations, saved in Troj format.
    :param annotations: Annotations dictionary in Troj format.
    :param corruptions: A list of corruptions.
    :param severities: A list of severity levels.
    :param mode: One of either 'random' or 'all'. If random, each image is corrupted by randomly sampling a corruption and severity level.
    If all, every corruption and every severity is applied to each image one at a time and saved.
    :param include_original: Whether or not to include the original image in the new test set.
    :param verbose: Whether or not to print generation progress.
    :returns: Creates a new dataset at new_folder containing corrupt images.
    '''
    if mode.lower()=='random':
        corrupt_annotations = corrupt_od_dset_random(original_folder, new_folder, annot_save_loc, annotations, corruptions, severities, include_original, verbose)
    if mode.lower()=='all':
        corrupt_annotations = corrupt_od_dset_all(original_folder, new_folder, annot_save_loc, annotations, corruptions, severities, include_original, verbose)
    return corrupt_annotations



import os
import PIL


def DatasetFormatConverterLabelled(dset, dset_dir, split, label_id_map, extension='.jpeg', verbose=True):
    '''
    Takes in a PyTorch dataset and converts it to the Imagenet folder structure.

    :param dset: Pytorch dataset object.
    :param dset_dir: Directory to save the new dataset.
    :param split: Split name (i.e train, test, val)
    :param label_id_map: a dictionary where the keys are the names of the classes and the values are the assigned integer labels.
    :param extension: File extension to save the data with (i.e)
    :param verbose: Whether or not to print progress.
    :return: A dataset split saved at dset_dir/split.
    '''
    rev_lab_id_map = dict(zip(label_id_map.values(), label_id_map.keys()))
    if os.path.exists(dset_dir) == False:
        os.mkdir(dset_dir)
    split_dir = os.path.join(dset_dir, split)
    os.mkdir(split_dir)
    for class_name in list(label_id_map.keys()):
        class_dir = os.path.join(split_dir, class_name)
        os.mkdir(class_dir)
    tracker = 0
    for data_lab_pair in dset:
        if verbose:
            if tracker%100==0:
                print('{}/{} converted'.format(tracker, len(dset)))
        img = data_lab_pair[0]
        label = data_lab_pair[1]
        class_name = rev_lab_id_map[label]
        class_dir = os.path.join(split_dir, class_name)
        save_file = '{}_{}{}'.format(label, tracker, extension)
        save_loc = os.path.join(class_dir, save_file)
        img.save(save_loc)
        tracker += 1


def DatasetFormatConverterUnlabelled(dset, dset_dir, split='unlabeled',extension='.jpeg', verbose=True):
    '''
    Takes in a PyTorch dataset and converts it to the Imagenet folder structure, however there are no class folders.

    :param dset: Pytorch dataset object.
    :param dset_dir: Directory to save the new dataset.
    :param split: Split name (i.e train, test, val)
    :param extension: File extension to save the data with (i.e)
    :param verbose: Whether or not to print progress.
    :return: A dataset split saved at dset_dir/split.
    '''
    if os.path.exists(dset_dir) == False:
        os.mkdir(dset_dir)
    split_dir = os.path.join(dset_dir, split)
    os.mkdir(split_dir)
    tracker = 0
    for data_lab_pair in dset:
        if verbose:
            if tracker%100==0:
                print('{}/{} converted'.format(tracker, len(dset)))
        img = data_lab_pair[0]
        save_file = 'unlab_{}{}'.format(tracker, extension)
        save_loc = os.path.join(split_dir, save_file)
        img.save(save_loc)
        tracker += 1


# -*- coding: utf-8 -*-

'''
From https://github.com/bethgelab/imagecorruptions, included in Troj package for modification and consistency.
'''

import numpy as np
import math
from PIL import Image

# /////////////// Corruption Helpers ///////////////

import skimage as sk
from skimage.filters import gaussian
from io import BytesIO
import cv2
from scipy.ndimage import zoom as scizoom
from scipy.ndimage.interpolation import map_coordinates
import warnings
import os
from pkg_resources import resource_filename
from numba import njit, prange



def disk(radius, alias_blur=0.1, dtype=np.float32):
    if radius <= 8:
        L = np.arange(-8, 8 + 1)
        ksize = (3, 3)
    else:
        L = np.arange(-radius, radius + 1)
        ksize = (5, 5)
    X, Y = np.meshgrid(L, L)
    aliased_disk = np.array((X ** 2 + Y ** 2) <= radius ** 2, dtype=dtype)
    aliased_disk /= np.sum(aliased_disk)

    # supersample disk to antialias
    return cv2.GaussianBlur(aliased_disk, ksize=ksize, sigmaX=alias_blur)




# modification of https://github.com/FLHerne/mapgen/blob/master/diamondsquare.py
def plasma_fractal(mapsize=256, wibbledecay=3):
    """
    Generate a heightmap using diamond-square algorithm.
    Return square 2d array, side length 'mapsize', of floats in range 0-255.
    'mapsize' must be a power of two.
    """
    assert (mapsize & (mapsize - 1) == 0)
    maparray = np.empty((mapsize, mapsize), dtype=np.float_)
    maparray[0, 0] = 0
    stepsize = mapsize
    wibble = 100

    def wibbledmean(array):
        return array / 4 + wibble * np.random.uniform(-wibble, wibble,
                                                      array.shape)

    def fillsquares():
        """For each square of points stepsize apart,
           calculate middle value as mean of points + wibble"""
        cornerref = maparray[0:mapsize:stepsize, 0:mapsize:stepsize]
        squareaccum = cornerref + np.roll(cornerref, shift=-1, axis=0)
        squareaccum += np.roll(squareaccum, shift=-1, axis=1)
        maparray[stepsize // 2:mapsize:stepsize,
        stepsize // 2:mapsize:stepsize] = wibbledmean(squareaccum)

    def filldiamonds():
        """For each diamond of points stepsize apart,
           calculate middle value as mean of points + wibble"""
        mapsize = maparray.shape[0]
        drgrid = maparray[stepsize // 2:mapsize:stepsize,
                 stepsize // 2:mapsize:stepsize]
        ulgrid = maparray[0:mapsize:stepsize, 0:mapsize:stepsize]
        ldrsum = drgrid + np.roll(drgrid, 1, axis=0)
        lulsum = ulgrid + np.roll(ulgrid, -1, axis=1)
        ltsum = ldrsum + lulsum
        maparray[0:mapsize:stepsize,
        stepsize // 2:mapsize:stepsize] = wibbledmean(ltsum)
        tdrsum = drgrid + np.roll(drgrid, 1, axis=1)
        tulsum = ulgrid + np.roll(ulgrid, -1, axis=0)
        ttsum = tdrsum + tulsum
        maparray[stepsize // 2:mapsize:stepsize,
        0:mapsize:stepsize] = wibbledmean(ttsum)

    while stepsize >= 2:
        fillsquares()
        filldiamonds()
        stepsize //= 2
        wibble /= wibbledecay

    maparray -= maparray.min()
    return maparray / maparray.max()


def clipped_zoom(img, zoom_factor):
    # clipping along the width dimension:
    ch0 = int(np.ceil(img.shape[0] / float(zoom_factor)))
    top0 = (img.shape[0] - ch0) // 2

    # clipping along the height dimension:
    ch1 = int(np.ceil(img.shape[1] / float(zoom_factor)))
    top1 = (img.shape[1] - ch1) // 2

    img = scizoom(img[top0:top0 + ch0, top1:top1 + ch1],
                  (zoom_factor, zoom_factor, 1), order=1)

    return img

def getOptimalKernelWidth1D(radius, sigma):
    return radius * 2 + 1

def gauss_function(x, mean, sigma):
    return (np.exp(- x**2 / (2 * (sigma**2)))) / (np.sqrt(2 * np.pi) * sigma)

def getMotionBlurKernel(width, sigma):
    k = gauss_function(np.arange(width), 0, sigma)
    Z = np.sum(k)
    return k/Z

def shift(image, dx, dy):
    if(dx < 0):
        shifted = np.roll(image, shift=image.shape[1]+dx, axis=1)
        shifted[:,dx:] = shifted[:,dx-1:dx]
    elif(dx > 0):
        shifted = np.roll(image, shift=dx, axis=1)
        shifted[:,:dx] = shifted[:,dx:dx+1]
    else:
        shifted = image

    if(dy < 0):
        shifted = np.roll(shifted, shift=image.shape[0]+dy, axis=0)
        shifted[dy:,:] = shifted[dy-1:dy,:]
    elif(dy > 0):
        shifted = np.roll(shifted, shift=dy, axis=0)
        shifted[:dy,:] = shifted[dy:dy+1,:]
    return shifted

def _motion_blur(x, radius, sigma, angle):
    width = getOptimalKernelWidth1D(radius, sigma)
    kernel = getMotionBlurKernel(width, sigma)
    point = (width * np.sin(np.deg2rad(angle)), width * np.cos(np.deg2rad(angle)))
    hypot = math.hypot(point[0], point[1])

    blurred = np.zeros_like(x, dtype=np.float32)
    for i in range(width):
        dy = -math.ceil(((i*point[0]) / hypot) - 0.5)
        dx = -math.ceil(((i*point[1]) / hypot) - 0.5)
        if (np.abs(dy) >= x.shape[0] or np.abs(dx) >= x.shape[1]):
            # simulated motion exceeded image borders
            break
        shifted = shift(x, dx, dy)
        blurred = blurred + kernel[i] * shifted
    return blurred

# Numba nopython compilation to shuffle_pixles
@njit()
def _shuffle_pixels_njit_glass_blur(d0,d1,x,c):

    # locally shuffle pixels
    for i in range(c[2]):
        for h in range(d0 - c[1], c[1], -1):
            for w in range(d1 - c[1], c[1], -1):
                dx, dy = np.random.randint(-c[1], c[1], size=(2,))
                h_prime, w_prime = h + dy, w + dx
                # swap
                x[h, w], x[h_prime, w_prime] = x[h_prime, w_prime], x[h, w]
    return x

# /////////////// End Corruption Helpers ///////////////


# /////////////// Corruptions ///////////////

def gaussian_noise(x, severity=1):
    c = [.08, .12, 0.18, 0.26, 0.38][severity - 1]

    x = np.array(x) / 255.
    return np.clip(x + np.random.normal(size=x.shape, scale=c), 0, 1) * 255


def shot_noise(x, severity=1):
    c = [60, 25, 12, 5, 3][severity - 1]

    x = np.array(x) / 255.
    return np.clip(np.random.poisson(x * c) / float(c), 0, 1) * 255


def impulse_noise(x, severity=1):
    c = [.03, .06, .09, 0.17, 0.27][severity - 1]

    x = sk.util.random_noise(np.array(x) / 255., mode='s&p', amount=c)
    return np.clip(x, 0, 1) * 255


def speckle_noise(x, severity=1):
    c = [.15, .2, 0.35, 0.45, 0.6][severity - 1]

    x = np.array(x) / 255.
    return np.clip(x + x * np.random.normal(size=x.shape, scale=c), 0, 1) * 255


def gaussian_blur(x, severity=1):
    c = [1, 2, 3, 4, 6][severity - 1]

    x = gaussian(np.array(x) / 255., sigma=c, multichannel=True)
    return np.clip(x, 0, 1) * 255


def glass_blur(x, severity=1):
    # sigma, max_delta, iterations
    c = [(0.7, 1, 2), (0.9, 2, 1), (1, 2, 3), (1.1, 3, 2), (1.5, 4, 2)][
        severity - 1]

    x = np.uint8(
        gaussian(np.array(x) / 255., sigma=c[0], multichannel=True) * 255)

    x = _shuffle_pixels_njit_glass_blur(np.array(x).shape[0],np.array(x).shape[1],x,c)

    return np.clip(gaussian(x / 255., sigma=c[0], multichannel=True), 0,
                   1) * 255

def defocus_blur(x, severity=1):
    c = [(3, 0.1), (4, 0.5), (6, 0.5), (8, 0.5), (10, 0.5)][severity - 1]

    x = np.array(x) / 255.
    kernel = disk(radius=c[0], alias_blur=c[1])

    channels = []
    if len(x.shape) < 3 or x.shape[2] < 3:
        channels = np.array(cv2.filter2D(x, -1, kernel))
    else:
        for d in range(3):
            channels.append(cv2.filter2D(x[:, :, d], -1, kernel))
        channels = np.array(channels).transpose((1, 2, 0))

    return np.clip(channels, 0, 1) * 255


def motion_blur(x, severity=1):
    shape = np.array(x).shape
    c = [(10, 3), (15, 5), (15, 8), (15, 12), (20, 15)][severity - 1]
    x = np.array(x)

    angle = np.random.uniform(-45, 45)
    x = _motion_blur(x, radius=c[0], sigma=c[1], angle=angle)

    if len(x.shape) < 3 or x.shape[2] < 3:
        gray = np.clip(np.array(x).transpose((0, 1)), 0, 255)
        if len(shape) >= 3 or shape[2] >=3:
            return np.stack([gray, gray, gray], axis=2)
        else:
            return gray
    else:
        return np.clip(x, 0, 255)


def zoom_blur(x, severity=1):
    c = [np.arange(1, 1.11, 0.01),
         np.arange(1, 1.16, 0.01),
         np.arange(1, 1.21, 0.02),
         np.arange(1, 1.26, 0.02),
         np.arange(1, 1.31, 0.03)][severity - 1]

    x = (np.array(x) / 255.).astype(np.float32)
    out = np.zeros_like(x)

    set_exception = False
    for zoom_factor in c:
        if len(x.shape) < 3 or x.shape[2] < 3:
            x_channels = np.array([x, x, x]).transpose((1, 2, 0))
            zoom_layer = clipped_zoom(x_channels, zoom_factor)
            zoom_layer = zoom_layer[:x.shape[0], :x.shape[1], 0]
        else:
            zoom_layer = clipped_zoom(x, zoom_factor)
            zoom_layer = zoom_layer[:x.shape[0], :x.shape[1], :]

        try:
            out += zoom_layer
        except ValueError:
            set_exception = True
            out[:zoom_layer.shape[0], :zoom_layer.shape[1]] += zoom_layer

    if set_exception:
        print('ValueError for zoom blur, Exception handling')
    x = (x + out) / (len(c) + 1)
    return np.clip(x, 0, 1) * 255


def next_power_of_2(x):
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


def fog(x, severity=1):
    c = [(1.5, 2), (2., 2), (2.5, 1.7), (2.5, 1.5), (3., 1.4)][severity - 1]

    shape = np.array(x).shape
    max_side = np.max(shape)
    map_size = next_power_of_2(int(max_side))

    x = np.array(x) / 255.
    max_val = x.max()

    x_shape = np.array(x).shape
    if len(x_shape) < 3 or x_shape[2] < 3:
        x += c[0] * plasma_fractal(mapsize=map_size, wibbledecay=c[1])[
                    :shape[0], :shape[1]]
    else:
        x += c[0] * \
             plasma_fractal(mapsize=map_size, wibbledecay=c[1])[:shape[0],
             :shape[1]][..., np.newaxis]
    return np.clip(x * max_val / (max_val + c[0]), 0, 1) * 255


def frost(x, severity=1):
    c = [(1, 0.4),
         (0.8, 0.6),
         (0.7, 0.7),
         (0.65, 0.7),
         (0.6, 0.75)][severity - 1]

    idx = np.random.randint(5)
    filename = [resource_filename(__name__, './frost/frost1.png'),
                resource_filename(__name__, './frost/frost2.png'),
                resource_filename(__name__, './frost/frost3.png'),
                resource_filename(__name__, './frost/frost4.jpg'),
                resource_filename(__name__, './frost/frost5.jpg'),
                resource_filename(__name__, './frost/frost6.jpg')][idx]
    frost = cv2.imread(filename)
    frost_shape = frost.shape
    x_shape = np.array(x).shape

    # resize the frost image so it fits to the image dimensions
    scaling_factor = 1
    if frost_shape[0] >= x_shape[0] and frost_shape[1] >= x_shape[1]:
        scaling_factor = 1
    elif frost_shape[0] < x_shape[0] and frost_shape[1] >= x_shape[1]:
        scaling_factor = x_shape[0] / frost_shape[0]
    elif frost_shape[0] >= x_shape[0] and frost_shape[1] < x_shape[1]:
        scaling_factor = x_shape[1] / frost_shape[1]
    elif frost_shape[0] < x_shape[0] and frost_shape[1] < x_shape[
        1]:  # If both dims are too small, pick the bigger scaling factor
        scaling_factor_0 = x_shape[0] / frost_shape[0]
        scaling_factor_1 = x_shape[1] / frost_shape[1]
        scaling_factor = np.maximum(scaling_factor_0, scaling_factor_1)

    scaling_factor *= 1.1
    new_shape = (int(np.ceil(frost_shape[1] * scaling_factor)),
                 int(np.ceil(frost_shape[0] * scaling_factor)))
    frost_rescaled = cv2.resize(frost, dsize=new_shape,
                                interpolation=cv2.INTER_CUBIC)

    # randomly crop
    x_start, y_start = np.random.randint(0, frost_rescaled.shape[0] - x_shape[
        0]), np.random.randint(0, frost_rescaled.shape[1] - x_shape[1])

    if len(x_shape) < 3 or x_shape[2] < 3:
        frost_rescaled = frost_rescaled[x_start:x_start + x_shape[0],
                         y_start:y_start + x_shape[1]]
        frost_rescaled = rgb2gray(frost_rescaled)
    else:
        frost_rescaled = frost_rescaled[x_start:x_start + x_shape[0],
                         y_start:y_start + x_shape[1]][..., [2, 1, 0]]
    return np.clip(c[0] * np.array(x) + c[1] * frost_rescaled, 0, 255)


def rgb2gray(rgb):
    return np.dot(rgb[..., :3], [0.2989, 0.5870, 0.1140])


def snow(x, severity=1):
    c = [(0.1, 0.3, 3, 0.5, 10, 4, 0.8),
         (0.2, 0.3, 2, 0.5, 12, 4, 0.7),
         (0.55, 0.3, 4, 0.9, 12, 8, 0.7),
         (0.55, 0.3, 4.5, 0.85, 12, 8, 0.65),
         (0.55, 0.3, 2.5, 0.85, 12, 12, 0.55)][severity - 1]

    x = np.array(x, dtype=np.float32) / 255.
    snow_layer = np.random.normal(size=x.shape[:2], loc=c[0],
                                  scale=c[1])  # [:2] for monochrome

    snow_layer = clipped_zoom(snow_layer[..., np.newaxis], c[2])
    snow_layer[snow_layer < c[3]] = 0

    snow_layer = np.clip(snow_layer.squeeze(), 0, 1)


    snow_layer = _motion_blur(snow_layer, radius=c[4], sigma=c[5], angle=np.random.uniform(-135, -45))

    # The snow layer is rounded and cropped to the img dims
    snow_layer = np.round(snow_layer * 255).astype(np.uint8) / 255.
    snow_layer = snow_layer[..., np.newaxis]
    snow_layer = snow_layer[:x.shape[0], :x.shape[1], :]

    if len(x.shape) < 3 or x.shape[2] < 3:
        x = c[6] * x + (1 - c[6]) * np.maximum(x, x.reshape(x.shape[0],
                                                            x.shape[
                                                                1]) * 1.5 + 0.5)
        snow_layer = snow_layer.squeeze(-1)
    else:
        x = c[6] * x + (1 - c[6]) * np.maximum(x, cv2.cvtColor(x,
                                                               cv2.COLOR_RGB2GRAY).reshape(
            x.shape[0], x.shape[1], 1) * 1.5 + 0.5)
    try:
        return np.clip(x + snow_layer + np.rot90(snow_layer, k=2), 0, 1) * 255
    except ValueError:
        print('ValueError for Snow, Exception handling')
        x[:snow_layer.shape[0], :snow_layer.shape[1]] += snow_layer + np.rot90(
            snow_layer, k=2)
        return np.clip(x, 0, 1) * 255



def spatter(x, severity=1):
    c = [(0.65, 0.3, 4, 0.69, 0.6, 0),
         (0.65, 0.3, 3, 0.68, 0.6, 0),
         (0.65, 0.3, 2, 0.68, 0.5, 0),
         (0.65, 0.3, 1, 0.65, 1.5, 1),
         (0.67, 0.4, 1, 0.65, 1.5, 1)][severity - 1]
    x_PIL = x
    x = np.array(x, dtype=np.float32) / 255.

    liquid_layer = np.random.normal(size=x.shape[:2], loc=c[0], scale=c[1])

    liquid_layer = gaussian(liquid_layer, sigma=c[2])
    liquid_layer[liquid_layer < c[3]] = 0
    if c[5] == 0:
        liquid_layer = (liquid_layer * 255).astype(np.uint8)
        dist = 255 - cv2.Canny(liquid_layer, 50, 150)
        dist = cv2.distanceTransform(dist, cv2.DIST_L2, 5)
        _, dist = cv2.threshold(dist, 20, 20, cv2.THRESH_TRUNC)
        dist = cv2.blur(dist, (3, 3)).astype(np.uint8)
        dist = cv2.equalizeHist(dist)
        ker = np.array([[-2, -1, 0], [-1, 1, 1], [0, 1, 2]])
        dist = cv2.filter2D(dist, cv2.CV_8U, ker)
        dist = cv2.blur(dist, (3, 3)).astype(np.float32)

        m = cv2.cvtColor(liquid_layer * dist, cv2.COLOR_GRAY2BGRA)
        m /= np.max(m, axis=(0, 1))
        m *= c[4]
        # water is pale turqouise
        color = np.concatenate((175 / 255. * np.ones_like(m[..., :1]),
                                238 / 255. * np.ones_like(m[..., :1]),
                                238 / 255. * np.ones_like(m[..., :1])), axis=2)

        color = cv2.cvtColor(color, cv2.COLOR_BGR2BGRA)

        if len(x.shape) < 3 or x.shape[2] < 3:
            add_spatter_color = cv2.cvtColor(np.clip(m * color, 0, 1),
                                             cv2.COLOR_BGRA2BGR)
            add_spatter_gray = rgb2gray(add_spatter_color)

            return np.clip(x + add_spatter_gray, 0, 1) * 255

        else:

            x = cv2.cvtColor(x, cv2.COLOR_BGR2BGRA)

            return cv2.cvtColor(np.clip(x + m * color, 0, 1),
                                cv2.COLOR_BGRA2BGR) * 255
    else:
        m = np.where(liquid_layer > c[3], 1, 0)
        m = gaussian(m.astype(np.float32), sigma=c[4])
        m[m < 0.8] = 0

        x_rgb = np.array(x_PIL.convert('RGB'))

        # mud brown
        color = np.concatenate((63 / 255. * np.ones_like(x_rgb[..., :1]),
                                42 / 255. * np.ones_like(x_rgb[..., :1]),
                                20 / 255. * np.ones_like(x_rgb[..., :1])),
                               axis=2)
        color *= m[..., np.newaxis]
        if len(x.shape) < 3 or x.shape[2] < 3:
            x *= (1 - m)
            return np.clip(x + rgb2gray(color), 0, 1) * 255

        else:
            x *= (1 - m[..., np.newaxis])
            return np.clip(x + color, 0, 1) * 255


def contrast(x, severity=1):
    c = [0.4, .3, .2, .1, .05][severity - 1]

    x = np.array(x) / 255.
    means = np.mean(x, axis=(0, 1), keepdims=True)
    return np.clip((x - means) * c + means, 0, 1) * 255


def brightness(x, severity=1):
    c = [.1, .2, .3, .4, .5][severity - 1]

    x = np.array(x) / 255.

    if len(x.shape) < 3 or x.shape[2] < 3:
        x = np.clip(x + c, 0, 1)
    else:
        x = sk.color.rgb2hsv(x)
        x[:, :, 2] = np.clip(x[:, :, 2] + c, 0, 1)
        x = sk.color.hsv2rgb(x)

    return np.clip(x, 0, 1) * 255


def saturate(x, severity=1):
    c = [(0.3, 0), (0.1, 0), (2, 0), (5, 0.1), (20, 0.2)][severity - 1]

    x = np.array(x) / 255.

    gray_scale = False
    if len(x.shape) < 3 or x.shape[2] < 3:
        x = np.array([x, x, x]).transpose((1, 2, 0))
        gray_scale = True
    x = sk.color.rgb2hsv(x)
    x[:, :, 1] = np.clip(x[:, :, 1] * c[0] + c[1], 0, 1)
    x = sk.color.hsv2rgb(x)
    if gray_scale:
        x = x[:, :, 0]

    return np.clip(x, 0, 1) * 255


def jpeg_compression(x, severity=1):
    c = [25, 18, 15, 10, 7][severity - 1]

    output = BytesIO()
    gray_scale = False
    if x.mode != 'RGB':
        gray_scale = True
        x = x.convert('RGB')
    x.save(output, 'JPEG', quality=c)
    x = Image.open(output)
    if gray_scale:
        x = x.convert('L')

    return x


def pixelate(x, severity=1):
    c = [0.6, 0.5, 0.4, 0.3, 0.25][severity - 1]

    x_shape = np.array(x).shape

    x = x.resize((int(x_shape[1] * c), int(x_shape[0] * c)), Image.BOX)

    x = x.resize((x_shape[1], x_shape[0]), Image.NEAREST)

    return x


# mod of https://gist.github.com/erniejunior/601cdf56d2b424757de5
def elastic_transform(image, severity=1):
    image = np.array(image, dtype=np.float32) / 255.
    shape = image.shape
    shape_size = shape[:2]

    sigma = np.array(shape_size) * 0.01
    alpha = [250 * 0.05, 250 * 0.065, 250 * 0.085, 250 * 0.1, 250 * 0.12][
        severity - 1]
    max_dx = shape[0] * 0.005
    max_dy = shape[0] * 0.005

    dx = (gaussian(np.random.uniform(-max_dx, max_dx, size=shape[:2]),
                   sigma, mode='reflect', truncate=3) * alpha).astype(
        np.float32)
    dy = (gaussian(np.random.uniform(-max_dy, max_dy, size=shape[:2]),
                   sigma, mode='reflect', truncate=3) * alpha).astype(
        np.float32)

    if len(image.shape) < 3 or image.shape[2] < 3:
        x, y = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]))
        indices = np.reshape(y + dy, (-1, 1)), np.reshape(x + dx, (-1, 1))
    else:
        dx, dy = dx[..., np.newaxis], dy[..., np.newaxis]
        x, y, z = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]),
                              np.arange(shape[2]))
        indices = np.reshape(y + dy, (-1, 1)), np.reshape(x + dx,
                                                          (-1, 1)), np.reshape(
            z, (-1, 1))
    return np.clip(
        map_coordinates(image, indices, order=1, mode='reflect').reshape(
            shape), 0, 1) * 255

# /////////////// End Corruptions ///////////////



def log_to_dataframe(dataframe, index, log_dict):
    """
    Logs data to dataframe.

    :param dataframe: Dataframe to log to.
    :param index: The indices to log to.
    :param log_dict: The dictionary containing data to log
    :return: updated dataframe.
    """
    for key in list(log_dict.keys()):
        if key not in dataframe.columns:
            dataframe[key] = ""
            dataframe[key].astype("object")
        dataframe.loc[index, key] = log_dict[key]
    return dataframe



import pandas as pd
import json
import numpy as np
import os
import cv2
import torch
import torch.utils.data
from torchvision import transforms
from PIL import Image


class TrojClassificationDataLoader:
    def __init__(
        self,
        dataframe,
        root_folder,
        transforms=[None],
        channel_first=True,
        return_idx=True,
    ):
        """
        :param dataframe: A troj dataframe.
        :param root_folder: The root folder containing the images.
        :param transforms: A (list of) function(s) to be applied to the images during loading.
        :param channel_first: Whether or not to put the channels first. Set to false if this is handled by transforms.
        """
        if type(transforms) is not list:
            transforms = [transforms]

        self.dataset_meta = {
            "root_folder": root_folder,
            "transforms": list(map(self._get_transforms, transforms)),
        }
        self.dataframe = dataframe
        self.transforms = transforms
        self.root_folder = root_folder
        self.index_list = self.dataframe.index.tolist()
        self.channel_first = channel_first
        self.return_idx = return_idx

    def _get_transforms(self, n):
        return n.__name__

    def __len__(self):
        return len(set(self.dataframe["file_name"]))

    def __getitem__(self, index):
        """

        :param index: Index to retrieve item from.
        :return: image and corresponding label, along with the dataframe index if self.return_idx = True
        """
        # print(self.dataframe.loc[index]["file_name"].compute())
        index = self.index_list[index]
        example_row = self.dataframe.loc[index]
        file_name = example_row["file_name"]
        annotation = example_row["label"]
        class_folder = example_row["class_name"]
        stage_folder = example_row["stage"]
        relative_path = (
            self.root_folder + "/" + stage_folder + "/" + class_folder + "/" + file_name
        )

        img = cv2.imread(relative_path)
        if self.transforms != None:
            for transform in self.transforms:
                img = transform(img)
        # find way to make work with troj-transforms?
        if self.channel_first is True:
            img = np.moveaxis(img, -1, 0)
        if self.return_idx:
            return img, annotation, index
        else:
            return img, annotation


class TrojClassificationDataset:
    def __init__(self):
        """
        We initialize the dataframe as None so that if the client does not wish to use our methods they can
        create their own dataframe and pass it to this class. We just specify that the dataframe
        must have the expected basic columns for our methods to function properly.
        """
        self.dataframe = None
        # saves the main folder to look in to recreate the relative links to the images for loading
        self.root_folder = None
        # saves whether or not the data is in coco format or imagenet folders
        self.data_structure = None

    def CreateDF(self, folder_path):
        """
        :param folder_path: A folder path which contains subfolders with the structure
        folder_path/stage_folder/class_folder/img.format
        :return: creates dataframe
        """

        try:
            # if no annotations are included, the annotation information is embedded in the folder structure
            # load it in to dataframe
            accumulator = 0
            out_dict = dict()
            class_list = list()
            for root, d_names, f_names in sorted(
                os.walk(os.path.normpath(folder_path))
            ):
                if f_names != []:
                    # replace the double slash with forward slash for consistency
                    # root = os.path.join(root).replace("\\", "/")
                    # split out each part of the root
                    root = os.path.realpath(root)
                    class_name = root.split(os.path.sep)[-1]
                    stage = root.split(os.path.sep)[-2]
                    base = root.split(os.path.sep)[-3]
                    if class_name not in class_list:
                        class_list.append(class_name)
                    for i, f_name in enumerate(f_names):
                        i = i + accumulator
                        out_dict[i] = {
                            "stage": stage,
                            "class_name": class_name,
                            "file_name": f_name,
                            "label": class_list.index(class_name),
                        }
                        # the accumulator tracks the index of the dictionary across folder loops
                    accumulator = i + 1
            df = pd.DataFrame(out_dict).transpose()
            self.data_structure = "imagenet"
        except Exception as ex:
            print(ex)
            print("Creating dataframe from imagefolders failed!")

        self.dataframe = df
        self.root_folder = folder_path

    def SaveDF(self, path, **kwargs):
        """
        :param path: the save path. This can be a globstring.
        :param kwargs: args for to_csv
        :return:
        """
        self.dataframe.to_csv(path, **kwargs)


def convert_coco_to_cartesian(bbox):
    """
    Converts coco style bounding box coordinates to cartesian coordinates.
    :param bbox:
    :return:
    """
    new_bbox = [bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]]
    return new_bbox


def convert_coco_json_to_dict(filename, stage, convert_bbox=True):
    """
    Converts a coco-style JSON to a dicionary.
    :param filename: JSON file
    :param stage: a name for the stage such as train, test, validation, etc.
    :param convert_bbox: Whether or not to convert from coco format to cartestian bounding boxes.
    :return: A dictionary with the image id as the key where the value is a dictionary containing the annotation information.
    """
    s = json.load(open(filename, "r"))
    annots = s["annotations"]
    ims = s["images"]
    data_dict = {}
    for im in ims:
        im_dict = im
        file_name = im_dict["file_name"]
        im_id = im_dict["id"]
        temp_dict = {}
        temp_dict["stage"] = stage
        temp_dict["file_name"] = file_name
        temp_dict["boxes"] = []
        temp_dict["labels"] = []

        data_dict[im_id] = temp_dict

    for annotations in annots:
        im_id = annotations["image_id"]
        cur_subdict = data_dict[im_id]
        coco_box = [annotations["bbox"][i] for i in range(len(annotations["bbox"]))]
        if convert_bbox:
            coco_box = convert_coco_to_cartesian(coco_box)
        cur_subdict["boxes"].append(coco_box)
        cur_subdict["labels"].append(annotations["category_id"])
        data_dict[im_id] = cur_subdict

    return data_dict


class TrojODDataset:
    def __init__(self):
        self.dataframe = None

    def CreateODDF(self, image_folder, annotations_dict, convert_coords=True):
        """
        Creates a dataframe for object detection datasets.

        :param: image_folder: folder containing split subfolders
        :param: annotations_dict: a dictionary where the keys are the names of subfolders, and the values are file locations for the
        annotations. If a folder has no annotations, set value to None in dict.
        :param: convert_coords: whether or not to convert coco-style coordinates to cartesian coordinates
        """
        data_list = []

        for im_dir in os.listdir(image_folder):
            if im_dir not in list(annotations_dict.keys()):
                pass
            else:
                im_dir_path = os.path.join(image_folder, im_dir)
                if annotations_dict[im_dir] == None:
                    pass
                else:
                    temp_data = convert_coco_json_to_dict(
                        annotations_dict[im_dir], im_dir, convert_coords
                    )
                    data_list.append(temp_data)

        dall = {}
        for d in data_list:
            dall.update(d)
        df = pd.DataFrame(list(dall.values()))

        self.dataframe = df


class ODTrojDataLoader:
    def __init__(self, dataframe, root_folder, transforms=None):
        """
        Dataloader for object detection data.

        :param dataframe: Troj object detection dataframe.
        :param root_folder: The root image folder.
        :param transforms: Any transforms to apply to the images (corresponding box transformations not yet implemented).
        """
        self.dataframe = dataframe
        self.transforms = transforms
        self.root_folder = root_folder
        self.index_list = self.dataframe.index.tolist()
        self.dataset_meta = {
            "root_folder": str(root_folder),
        }

    def __len__(self):
        return len(set(self.dataframe["file_name"]))

    def __getitem__(self, index):
        """

        :param index: Item index
        :return: transformed images, bounding boxes, labels, and the dataframe index
        """
        index = self.index_list[index]
        example_row = self.dataframe.loc[index]
        file_name = example_row["file_name"]
        labels = example_row["labels"]
        boxes = example_row["boxes"]
        stage_folder = example_row["stage"]
        relative_path = self.root_folder + "/" + stage_folder + "/" + file_name

        img = Image.open(relative_path)
        boxes = torch.tensor(boxes)
        labels = torch.tensor(labels)
        if self.transforms == None:
            self.transforms = transforms.ToTensor()
            # apply transforms
        return self.transforms(img), boxes, labels, index


def BuildODBatchIterator(dataloader, batch_size, shuffle=True, device="cuda", **kwargs):
    """
    The Troj function for iterating over OD data given a dataloader.

    :param dataloader: Troj OD Dataloader instance
    :param batch_size: the size of batches to be returned
    :param shuffle: Whether or not to shuffle the data or keep it in the original order
    :param device: device to put the data on
    :param kwargs:
    :return: A batch iterator
    """

    def ODcollate_wrapper(batch, device=device):
        transposed_data = list(zip(*batch))
        inp = []
        dict_list = []
        for idx in range(len(transposed_data[1])):
            # inp.append(transposed_data[0][idx])
            inp.append(transposed_data[0][idx].to(device))
            temp_dict = {}
            temp_dict["labels"] = transposed_data[2][idx].to(device)
            temp_dict["boxes"] = transposed_data[1][idx].to(device)
            dict_list.append(temp_dict)

        idx = np.stack(transposed_data[3], 0)
        return (inp, dict_list, idx)

    loader = torch.utils.data.DataLoader(
        dataloader,
        batch_size=batch_size,
        shuffle=shuffle,
        collate_fn=ODcollate_wrapper,
        **kwargs
    )
    return loader


def BuildClassificationBatchIterator(
    dataloader, batch_size, shuffle=True, as_torch_tensor=False, **kwargs
):
    """
    :param dataloader: Troj Classification dataloader.
    :param batch_size: size of batches.
    :param shuffle: Whether or not to shuffle the data.
    :param kwargs:
    :return:
    """

    def collate_wrapper(batch):
        transposed_data = list(zip(*batch))
        inp = np.ascontiguousarray(np.stack(transposed_data[0], 0).astype(np.float32))
        tgt = np.stack(transposed_data[1], 0).astype(np.int64)
        if as_torch_tensor:
            inp = torch.from_numpy(inp)
            tgt = torch.from_numpy(tgt)
        if len(transposed_data) == 3:
            idx = np.stack(transposed_data[2], 0)
            return (inp, tgt, idx)
        else:
            return (inp, tgt)

    loader = torch.utils.data.DataLoader(
        dataloader,
        batch_size=batch_size,
        shuffle=shuffle,
        collate_fn=collate_wrapper,
        **kwargs
    )
    return loader



import torchvision.ops
import torch
import numpy as np
from .attack_utils import nms_reduction, get_objects_of_type, check_flip


class EvoDAttack:
    def __init__(
        self,
        model,
        attack_class,
        iterations=25,
        init_pop_size=4,
        thresh=0.05,
        max_val=0.1,
        init_pop_mean=0,
        init_pop_dev=0.05,
        num_offspring=2,
        mut_mean=0,
        mut_deviation=0.005,
        pop_reduction_factor=2,
        device="cuda",
        **kwargs
    ):
        """
        A simple blackbox attack for object detection algorithms using evolutionary strategies, intended as a measure of
        classifier robustness against a simple adversary. The user selects a class the want to try to fool. Given an image
        containing such a class, the algorithm finds the instance in the image of that class and tries to find noise which
        causes the classifier to miss that instance.

        :param model: Pytorch Object detection model. That is, any model which outputs predictions as a dictionary
        containing the keys [boxes, labels, scores] where boxes are the bounding boxes for
        predicted objects in the form [x,y, x+w, y+h] where w,h are the width and height of the bounding box. Labels are the
        predicted labels, and scores are the confidence of prediction, where each value is of the correct Pytorch dtype.
        :param attack_class: The class to perform the attack on.
        :param iterations: Number of generations for the attack.
        :param init_pop_size: Size of the initial population
        :param thresh: The maximum fitness for an individual is 0. The value thresh specifies a stopping bound by -thresh
        meaning that if an individual with such a fitness score is found, the algorithm stops.
        :param max_val: Maximum allowable perturbation
        :param init_pop_mean: Mean of the values in the initial population. (population sampled from Gaussian)
        :param init_pop_dev: Standard deviation of the initial population.
        :param num_offspring: Number of offspring at each timestep.
        :param mut_mean: Mean value of mutations applied to offspring
        :param mut_deviation: Deviation of mutation values.
        :param pop_reduction_factor: Factor to reduce the population by after each timestep.
        :param device: Device for tensors
        :param kwargs:
        """
        if type(model).__name__ == "PytorchObjectDetector":
            self.model = model.model
        else:
            self.model = model
        self.attack_class = attack_class
        self.init_pop_size = init_pop_size
        self.thresh = thresh
        self.max_val = max_val
        self.iterations = iterations
        self.init_pop_mean = init_pop_mean
        self.init_pop_dev = init_pop_dev
        self.num_offspring = num_offspring
        self.mut_mean = mut_mean
        self.mut_deviation = mut_deviation
        self.device = device
        self.pop_reduction_factor = pop_reduction_factor

    def get_weakest(self, predictions):
        """
        Takes in the predictions for the class we care about, and finds the weakest one

        :param predictions: Predictions of class we wish to fool
        :return: Dictionary containing the weakest sample
        """
        scores = predictions["scores"]
        weak_id = torch.argmin(scores)
        reduced_predictions = {
            "boxes": predictions["boxes"][weak_id],
            "labels": predictions["labels"][weak_id],
            "scores": predictions["scores"][weak_id],
        }
        return reduced_predictions

    def get_ground_truth(self, annots, wl_pred):
        """
        Finds the ground truth object for the weakest prediction of the model for the class we are attacking.

        :param annots: Labeled sample containing only instances of class we wish to perturb
        :param wl_pred: Model prediction of the class with the lowest score.
        :return: The ground truth annotation for the weakest prediction.
        """
        ious = torchvision.ops.box_iou(wl_pred["boxes"], annots["boxes"].squeeze(dim=1))
        gt_idx = torch.argmax(ious, dim=1)[0]
        true_gt = {"boxes": annots["boxes"][gt_idx], "labels": annots["labels"][gt_idx]}
        return true_gt

    def find_vulnerable(self, predicts, targ, thresh=0.05):
        """
        Finds the most vulnerable ground truth label.

        :param predicts: All model predictions on sample.
        :param targ: Targets for the sample.
        :param thresh: Thresh for NMS.
        :return: The most vulnerable ground truth instance.
        """
        obj_id = self.attack_class
        with torch.no_grad():
            red_preds = nms_reduction(predicts, thresh=thresh)
            object_preds = get_objects_of_type(predicts, obj_id, inc_scores=True)
            object_target = get_objects_of_type(targ, obj_id, inc_scores=False)
            if object_target["boxes"].shape[0] == 0:
                gt = None
            elif object_preds["boxes"].shape[0] == 0:
                gt = None
            else:
                weakest_link = self.get_weakest(object_preds)
                gt = self.get_ground_truth(object_target, weakest_link)
        return gt

    def get_prediction(self, gt, predictions):
        """
        Returns the label y ={1, -1} where y=1 if the label is the desired label, and -1 if the iou with
        the ground truth box is 0 (since the class has either flipped or is background). We also return the score, and the iou.
        :param gt: Ground truth object we are attacking.
        :param predictions: All model predictions
        :return: The IOU (intersection over union), the score, and the flip indication.
        """
        obj_type = gt["labels"][0]
        object_preds = get_objects_of_type(predictions, obj_type, inc_scores=True)
        if len(object_preds["labels"]) == 0:
            obj_iou = 0
            score = 0
            label = -1
        else:
            ious = torchvision.ops.box_iou(
                gt["boxes"], object_preds["boxes"].squeeze(dim=1)
            )
            comparison_tensor = torch.zeros_like(ious)
            if torch.equal(comparison_tensor, ious):
                obj_iou = 0
                score = 0
                label = -1
            else:
                associated_pred_id = torch.argmax(ious, dim=1)
                associated_pred = {
                    "boxes": object_preds["boxes"][associated_pred_id],
                    "labels": object_preds["labels"][associated_pred_id],
                    "scores": object_preds["scores"][associated_pred_id],
                }
                obj_iou = ious[0][associated_pred_id][0].item()
                label = 1
                score = object_preds["scores"][associated_pred_id][0].item()

        return obj_iou, score, label

    def generate_population(self, num_indiv, shape):
        """
        Generates initial population.

        :param num_indiv: Number of individuals to generate.
        :param shape: Shape of elements of population.
        :return: Initial population.
        """
        # generate initial population
        population = torch.nn.init.normal_(
            torch.zeros((num_indiv, shape[0], shape[1], shape[2]), requires_grad=False),
            mean=self.init_pop_mean,
            std=self.init_pop_dev,
        )
        return population

    def generate_offspring_asexual(self, population):
        """
        Generates offspring asexually (meaning no recombination between members of the population)

        :param population: Input population
        :return: New population with p_{i-1}*n individuals, where p_{i-1} is the number of individuals in the previous
        population and n is the number of offspring.
        """
        # generate offspring for recombination and add random mutations
        new_population = []
        for i in range(self.num_offspring - 1):
            new_population.append(population)
        new_population = torch.cat(new_population)
        mutations = torch.nn.init.normal_(
            torch.zeros_like(new_population, requires_grad=False),
            mean=self.mut_mean,
            std=self.mut_deviation,
        )
        new_population = new_population + mutations
        new_population = torch.clamp(new_population, -self.max_val, self.max_val)
        new_population = torch.cat([population, new_population])
        return new_population

    # inp is tensor of shape C, W, H (no batch dimension)
    def evaluate_fitness(self, inp, gt, population):
        """
        Evaluate fitness of population
        :param inp: Tensor of shape C, W, H (no batch dimension)
        :param gt: Ground truth annotation for vulnerable object.
        :param population: Current population
        :return: Fitness of population members and their returned flip indicators.
        """
        device = self.device
        # evaluate fitness of members of population
        self.model.eval()
        pop_fitness = []
        inp_tile = []
        for i in range(population.shape[0]):
            inp_tile.append(inp)
        inp_tile = torch.stack(inp_tile)
        perturbed_inp = inp_tile.cpu() + population
        perturbed_inp = [
            perturbed_inp[i].to(device) for i in range(perturbed_inp.shape[0])
        ]
        with torch.no_grad():
            model_preds = self.model(perturbed_inp)
        returned_labels = []
        for i in range(len(perturbed_inp)):
            iou, score, label = self.get_prediction(gt, model_preds[i])
            flip = check_flip(model_preds[i], gt, self.attack_class)
            if flip == 1:
                label = -1 * flip
            else:
                label = 1
            perturb = population[i]
            perturb = perturb.view(-1)
            pert_norm = torch.norm(perturb, p=np.inf)
            returned_labels.append(label)
            fitness = -1 * label - pert_norm - score * iou
            pop_fitness.append(fitness)
        return torch.Tensor(pop_fitness), torch.Tensor(returned_labels)

    def reduce_population(self, population, fitness, ret_lab):
        """

        Splits population based on reduction factor (reduces based on fitness)

        :param population: Current population
        :param fitness: Fitness of current population
        :param ret_lab: Flip indicators
        :return: Reduced population
        """
        sorted_fitness, indices = torch.sort(fitness, descending=True)
        remaining_pop = int(fitness.shape[0] / self.pop_reduction_factor)
        asc_to_desc = indices
        pop_by_fitness = population[asc_to_desc]
        lab_by_fitness = ret_lab[asc_to_desc]
        reduced_pop = pop_by_fitness[0:remaining_pop]
        reduced_lab = lab_by_fitness[0:remaining_pop]
        reduced_fitness = sorted_fitness[0:remaining_pop]
        return reduced_pop, reduced_lab, reduced_fitness

    def attack(self, x, y, verbose=True):
        """
        Perform an attack on a sample.

        :param x: a tensor of shape C W H with no batch dimension
        :param y:  y are the annotations for x converted to the appropriate format (bounding box as [x,y, x+w, y+h]).
        :param verbose: Print progress
        :return: The found perturbation, the vulnerable ground truth, and the initial model predictions.
        """
        pop_fits = []
        init_preds = self.model([x])[0]
        if y == None:
            y = init_preds
        gt = self.find_vulnerable(init_preds, y, self.attack_class)
        if gt == None:
            return None, None, None
        population = self.generate_population(self.init_pop_size, x.shape)
        for i in range(self.iterations):
            population = self.generate_offspring_asexual(population)
            fit, returned_labs = self.evaluate_fitness(x, gt, population)
            population, labs, red_fit = self.reduce_population(
                population, fit, returned_labs
            )
            av_fitness = torch.mean(red_fit)
            pop_fits.append(av_fitness.item())
            if verbose == True:
                print("average fitness at generation {}: ".format(i), av_fitness.item())
            if torch.max(fit) > self.thresh:
                break
        negated = torch.where(labs == 0, labs, torch.tensor([-1.0]).float()[0])
        flipped_locs = negated == -1
        flipped_idx = negated.nonzero()
        if flipped_idx.nelement() > 0:
            population = population[flipped_idx]

        outs = population[0]
        if len(outs.shape) > 3:
            outs = outs[0]
        # return av pop fitness
        return outs, gt, init_preds



from art import estimators
import numpy as np
import tensorflow as tf


"""


This class is our base class that child classes will inherit metadata collection (and other functions in the future) as well as inherits from
ART's base estimator class for instantiation and attack. 
We'll have a class for major frameworks that will inherit this class

"""


class TrojClassifier(estimators.BaseEstimator):
    """
    Collects and returns the model's metadata
    :return dict: returns a dictionary of metadata related to the classifier passed to the ART estimator class
    """

    def get_classifier_meta(self):
        return {
            "model": str(self._model),
            "model_name": str(self._model.__class__.__name__),
            "input_shape": str(self._input_shape),
            "num_classes": str(self.nb_classes),
            "loss_func": str(self._loss),
            "channels_first": str(self._channels_first),
        }


"""
Inherits from the troj base class and the ART tensorflow v1 classifier for tf1 support
Also implements our ComputeLoss function which will be called by each classifier, should implement as abstract method in base class
"""


class TrojTF1Classifier(estimators.classification.TensorFlowClassifier, TrojClassifier):
    from art.estimators.classification import TensorFlowClassifier

    """
    Compute loss computes the loss over a batch of target samples
    :param x: inputs
    :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
    the first dimension because of art bug.
    :param return_preds: bool to return model predictions with loss
    :return: Loss values and optionally predictions
    """

    def ComputeLoss(self, x, y, return_preds=True):
        try:
            import tensorflow.compat.v1 as tf
        except:
            import tensorflow as tf
        from numpy import array

        # this has got to be wrong
        shape = (y.size, self.nb_classes)
        one_hot = np.zeros(shape)
        losses = np.array([])
        for sample in x:
            image = np.expand_dims(sample, axis=0)
            rows = np.arange(y.size)
            print(rows)
            one_hot[rows, y] = 1.0
            print(one_hot)
            label = np.expand_dims(one_hot[0], axis=0)
            loss_val = self.compute_loss(x, one_hot)
            losses = np.append(losses, loss_val)

        if return_preds:
            preds = self.predict(x)
            return losses, preds
        else:
            return loss_val.numpy()


"""
Inherits from the troj base class and the ART tensorflow v1 classifier for tf1 support
Also implements our ComputeLoss function which will be called by each classifier, should implement as abstract method in base class
"""


class TrojKerasClassifier(estimators.classification.KerasClassifier, TrojClassifier):
    import tensorflow as tf

    tf.compat.v1.disable_eager_execution()

    """
    Compute loss computes the loss over a batch of target samples

    :param x: inputs
    :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
    the first dimension because of art bug.
    :param return_preds: bool to return model predictions with loss
    :param reduction: changes the reduction of the loss function to none 
    :return: Loss values and optionally predictions
    """

    def ComputeLoss(
        self, x, y, return_preds=True, reduction=tf.keras.losses.Reduction.NONE
    ):
        # import tensorflow.compat.v1 as tf
        import tensorflow as tf

        old_reduction = self.model.loss_functions[0]._get_reduction()
        self.model.loss_functions[0].reduction = reduction
        preds = self.predict(x)
        # hard_labels = np.argmax(preds, axis=1)
        # hard_labels_dtype = 'float32'
        # y = y.astype(hard_labels_dtype)
        loss_val = self.model.loss_functions[0](y, preds)
        self._loss.reduction = old_reduction
        # tf.compat.v1.enable_eager_execution()
        if return_preds:

            return loss_val.eval(session=tf.compat.v1.Session()), preds
        else:
            return loss_val.numpy()


"""
Inherits from the troj base class and the ART tensorflow v1 classifier for tf1 support
Also implements our ComputeLoss function which will be called by each classifier, should implement as abstract method in base class
"""


class TrojPytorchClassifier(
    estimators.classification.PyTorchClassifier, TrojClassifier
):
    """
    Compute loss computes the loss over a batch of target samples

    :param x: inputs
    :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
    the first dimension because of art bug.
    :param return_preds: bool to return model predictions with loss
    :return: Loss values and optionally predictions
    """

    def ComputeLoss(self, x, y, return_preds=True, reduction="none"):
        import torch

        old_reduction = self._loss.reduction
        self._loss.reduction = reduction
        preds = torch.tensor(self.predict(x))
        y = torch.tensor(y)
        loss_val = self._loss(preds, y)
        self._loss.reduction = old_reduction
        if return_preds:
            return loss_val.numpy(), preds.numpy()
        else:
            return loss_val.numpy()



import numpy as np
import json
from trojai.logger import logger

class TrojSession:
    def __init__(self):
        super().__init__()
        self.client = None
        self.logger = logger.Logger()
        self.graphs = None

    def run_name(self, name ):
        self.log({"run_name": name})

    def log_graph(self, html):
        self.log({"graphs": html})

    def log(self, in_dict):
        self.logger.log(in_dict)

    def create_project(self, project_name: str):
        return self.client.create_project(project_name)

    def create_dataset(self, project_name: str, dataset_name: str):
        return self.client.create_dataset(project_name, dataset_name)

    def reorient_df(self, dataframe):
        dataframe = json.loads(
            dataframe.to_json(orient="index")
        )
        return dataframe 

    """
    Drops rows with a null value in a column from the dataframe then reorients the dataframe to be returned
    :param dataframe: the dataframe
    :param drop_na: nulls should be dropped bool
    """
    def drop_empty_rows(self, dataframe, drop_na):
        if drop_na == True:
            dataframe = dataframe.dropna()
        
        json_processed_out = self.reorient_df(dataframe)
        
        return json_processed_out

    # def drop_rows_old(self, dataframe, drop_na):
    #     # this is a check to determine if the metadata colleection has been run
    #     # lambdas go off the dataframe key so we need to recreate 
    #     if "dataframe" not in dataframe:
    #         tmp = dataframe
    #         dataframe = {}
    #         dataframe = tmp
    #     # if the rows with nulls should be droppe
    #     if drop_na == True:
    #         dataframe["dataframe"] = dataframe["dataframe"].dropna()
        
    #     #  = self.reorient_df(dataframe)
    #     json_processed_out = json.loads(
    #         dataframe["dataframe"].to_json(orient="index")
    #     )

        
    #     return json_processed_out

    # def upload_dataframe(
    #     self, dataframe, project_name: str, dataset_name: str, drop_na=True
    # ):
    #     processed_df = self.drop_rows_old(dataframe, drop_na)
    #     jsonified_df = processed_df
    #     return self.client.upload_df_results(project_name, dataset_name, jsonified_df)

    """
    Uploads the logs collected by the logger

    :param project_name: string name of project that exusts on the platform
    :param dataset_name: a dataset that exists under the project  
    """
    def upload_logs(self, project_name: str, dataset_name: str, drop_na=True):
        logger_dict = self.logger.get_logger()
        processed_df = logger_dict
        processed_df["metadata"] = processed_df["metadata"][0]
        processed_df["dataframe"] = self.drop_empty_rows(logger_dict["dataframe"][0], drop_na)
        jsonified_df = processed_df
        print(jsonified_df)
        return self.client.upload_df_results(project_name, dataset_name, jsonified_df)

    '''
    Takes in the troj dataframe we've built over the attack loop as well as optionally the classifier, evaluator and dataloader
    for metadata collection. This function must be run before the upload_logs function for proper formatting.

    :param dataframe: the dataframe used to track the experiments and data
    :param classifier: the TrojClassifier object
    :param evaluator: The RobustnessEvaluator object
    :param dataloader: the dataloader object  
    :param tags: array of strings to tage the run 
    :return: dictionary of metadata
    '''
    def log_metadata(
        self, dataframe, classifier=None, evaluator=None, dataloader=None, tags=[]
    ):
        # also need to pull the log_dict from logger 
        classifier_meta = {}
        evaluator_meta = {}
        dataloader_meta = {}

        if classifier is not None:
            classifier_meta = classifier.get_classifier_meta()
        if evaluator is not None:
            evaluator_meta = evaluator.atk_meta
        if dataloader is not None:
            dataloader_meta = dataloader.dataset_meta
        if "prediction" in dataframe:
            dataframe["prediction"].replace("", np.nan, inplace=True)
        dataframe.dropna(inplace=True)
        self.logger = None
        self.logger = logger.Logger()
        out_dict = {
            "metadata": {
                "classifier_metadata": classifier_meta,
                "evaluator_metadata": evaluator_meta,
                "dataloader_metadata": dataloader_meta,
                "tags": str(tags),
            },
            "dataframe": dataframe,
        }
        self.log(out_dict)

        return out_dict



import pandas as pd
import numpy as np
import torch
from numpy import trapz

"""
TODO: run tests for each of these functions to increase coverage
"""

def MakePredAnnotBoxDataframes(df, adversarial=False):
    """
    Takes a troj OD dataframe and creates two new dataframes from it. The first dataframe contains the ground truth objects
    where each row is an instance of an object in an image. The second dataframe is the predicted objects, where each row is
    a single object in an image as well.
    :param df: Troj OD dataframe
    :param adversarial: If we are using the adversarial values or not.
    :return: dataframe containing ground truth objects and one containing predictions.
    """
    # makes two dataframes, where each row is a specific instance of a bounding box. Does this for both gt and preds
    gt_dicts_list = []
    pred_dicts_list = []
    ind_str = "{}_{}"
    if adversarial:
        prefix = "adversarial"
    else:
        prefix = "pred"
    for index, row in df.iterrows():
        im_name = row["file_name"]
        gt_labels = row["labels"]
        pred_labels = row[ind_str.format(prefix, "labels")]
        for idx in range(len(gt_labels)):
            label = row["labels"][idx]
            box = row["boxes"][idx]
            gt_dict = {"file_name": im_name, "label_id": label, "box": box}
            gt_dicts_list.append(gt_dict)
        for idx in range(len(pred_labels)):
            label = row[ind_str.format(prefix, "labels")][idx]
            box = row[ind_str.format(prefix, "boxes")][idx]
            score = row[ind_str.format(prefix, "scores")][idx]
            pred_dict = {
                "file_name": im_name,
                "pred_label_id": label,
                "pred_box": box,
                "pred_score": score,
            }
            pred_dicts_list.append(pred_dict)
    gt_box_df = pd.DataFrame(gt_dicts_list)
    pred_box_df = pd.DataFrame(pred_dicts_list)
    return gt_box_df, pred_box_df


def box_ious(gt_df, pred_df):
    """

    :param gt_df: ground truth dataframe from function MakePredAnnotBoxDataframes
    :param pred_df: Prediction dataframe from function MakePredAnnotBoxDataframes
    :return: returns both dataframes, where the pred_df has been modified to contain a column where each row in that column
    contains a number which is either -1 if the predicted object does not have a corresponding ground truth, or contains the
    index of the ground truth object in the ground truth dataframe. It also adds the maximum IOU value.
    """
    # get class ids
    class_ids = gt_df["label_id"].unique()
    # iterate over class ids
    for cls_id in class_ids:
        sub_pred_df = pred_df[pred_df["pred_label_id"] == cls_id]
        imgs = sub_pred_df["file_name"].unique()
        # iterate over images
        for img in imgs:
            # get image, get instances of classes in image.
            im_gt = gt_df[gt_df["file_name"] == img]
            im_gt = im_gt[im_gt["label_id"] == cls_id]
            im_gt_cls_boxes = np.asarray(im_gt["box"].tolist())
            pred_cls_boxes = np.asarray(
                sub_pred_df[sub_pred_df["file_name"] == img]["pred_box"].tolist()
            )
            # check for ground truth objects
            if im_gt_cls_boxes.shape[0] == 0:
                ious = torch.zeros(pred_cls_boxes.shape[0], 1)
                no_gt = True
            # if there are ground truth objects, compute the IOU between the objects and the predictions
            else:
                ious = torchvision.ops.box_iou(
                    torch.from_numpy(pred_cls_boxes), torch.from_numpy(im_gt_cls_boxes)
                )
                no_gt = False
            # get maximum ious for each object
            iou_max, indices = torch.max(ious, dim=1)
            indices = indices.numpy()
            iou_max = iou_max.numpy()
            inds = sub_pred_df[sub_pred_df["file_name"] == img].index
            pred_df.loc[inds, "max_iou"] = iou_max
            if no_gt == False:
                # save corresponding indices of most salient ground truth object
                ground_truth_inds = im_gt.iloc[indices].index
                pred_df.loc[inds, "gt_instance_idx"] = ground_truth_inds
                # go through gt indices in preds, for each gt_index keep only the maximum iou instance, set other index to -1
                for idx in ground_truth_inds:
                    comp_df = pred_df[pred_df["gt_instance_idx"] == idx]
                    comp_df = comp_df.sort_values("max_iou", ascending=False)
                    sorted_inds = comp_df.index
                    if len(sorted_inds) > 1:
                        pred_df.loc[sorted_inds[1:], "gt_instance_idx"] = (
                            -1 * torch.ones(len(sorted_inds[1:]), 1).numpy()
                        )
            else:
                # if no GT objects exist, set the indices to -1
                fp_markers = -1 * torch.ones(pred_cls_boxes.shape[0], 1).numpy()
                pred_df.loc[inds, "gt_instance_idx"] = fp_markers
            # pred_df = pred_df.loc[inds]['max_iou'] = iou_max
    return gt_df, pred_df


def compute_pr(tp, fp, gt_num):
    """
    Computes the precision and recall from the number of true positives, number of false positives, and the number of
    ground truth objects.
    :param tp: Number of true positves.
    :param fp: Number of false positives.
    :param gt_num: Number of ground truth objects.
    :return: precision, recall
    """
    # takes in # of true positives, # false positives, and total number of GT objects and returns the precision and recall
    recall = tp / gt_num
    precision = tp / (tp + fp)
    return precision, recall


def compute_interp_curve(precisions, recalls):
    """

    :param precisions: precisions array
    :param recalls: recalls array
    :return: interpolated PR curve.
    """
    curve_y = []
    for idx in range(recalls.shape[0]):
        p_interp = np.max(precisions[idx:])
        curve_y.append(p_interp)
    return curve_y


def ClassAP(gt_df, preds_df, iou_thresh=0.5):
    """
    Computes the average precision for each class according the PASCAL VOC criteria.
    :param gt_df: Ground truth dataframe.
    :param preds_df: predictions dataframe containing both the ground truth object column and the IOU column
    :param iou_thresh: the IOU threshold to compute the AP from.
    :return: A dictionary containing the class AP for each class at iou_thresh.
    """
    class_ap_dict = {}
    class_ids = gt_df["label_id"].unique()
    preds_copy = copy.deepcopy(preds_df)
    gt_copy = copy.deepcopy(gt_df)
    preds_copy.loc[(preds_copy["max_iou"] < iou_thresh), "gt_instance_idx"] = -1
    for cls_id in class_ids:
        tp_counter = 0
        fp_counter = 0
        precision = []
        recall = []
        sub_pred_df = preds_copy[preds_copy["pred_label_id"] == cls_id]
        sub_gt_df = gt_copy[gt_copy["label_id"] == cls_id]
        sub_pred_df = sub_pred_df.sort_values("pred_score", ascending=False)
        true_pos = sub_pred_df[sub_pred_df["gt_instance_idx"] != -1]
        num_gt = len(sub_gt_df)
        sub_preds_indices = sub_pred_df.index
        for idx in sub_preds_indices:
            if sub_pred_df.loc[idx]["gt_instance_idx"] == -1:
                fp_counter += 1
            else:
                tp_counter += 1
            prec, rec = compute_pr(tp_counter, fp_counter, num_gt)
            sub_pred_df.loc[idx, "accum_tp"] = tp_counter
            sub_pred_df.loc[idx, "accum_fp"] = fp_counter
            sub_pred_df.loc[idx, "precision"] = prec
            sub_pred_df.loc[idx, "recall"] = rec
        sorted_precisions = sub_pred_df["precision"].to_numpy()
        sorted_recalls = sub_pred_df["recall"].to_numpy()
        interp_curve = compute_interp_curve(sorted_precisions, sorted_recalls)
        ap = trapz(interp_curve, x=sorted_recalls)
        class_ap_dict[cls_id] = ap

    return class_ap_dict



from trojai.data import log_to_dataframe
from trojai.attacks.attack_utils import check_flip, get_class_ap, nms_pred_reduce
from trojai.metrics.metric_utils import get_pred_data
from trojai.attacks.BlackBoxODAttacks import EvoDAttack
import torch
import numpy as np


class BlackBoxODEvaluator:
    # different run form to classification
    def __init__(
        self,
        model,
        obj_class,
        loader,
        batch_iterator,
        df=None,
        device="cuda",
        iou_thresh=0.5,
        nms_thresh=0.05,
        return_nms=False,
        use_preds=False,
        verbose=True,
        **attkwargs
    ):
        """
        The class used for evaluating the robustness of an object detection algorithm against the simple blackbox attack.

        :param model:  Pytorch Object detection model. That is, any model which outputs predictions as a dictionary
        containing the keys [boxes, labels, scores] where boxes are the bounding boxes for
        predicted objects in the form [x,y, x+w, y+h] where w,h are the width and height of the bounding box. Labels are the
        predicted labels, and scores are the confidence of prediction, where each value is of the correct Pytorch dtype.
        :param obj_class: The class to perform the attack on.
        :param loader: Dataloader
        :param batch_iterator: Troj OD batch iterator
        :param df: Dataframe
        :param device: Deprecated
        :param iou_thresh: IOU threshold for a positive detection of an object
        :param nms_thresh: Threshold used when applying NMS.
        :param return_nms: When saving predictions to dataframe, whether or not apply NMS to the predictions first.
        :param use_preds: Whether or not to use the true annotations or annotations predicted by the model. #TODO implement
        :param verbose: Whether or not print info during attack.
        :param return_prc: Whether or not to return the precision recall curve.
        :param attkwargs: Arguments for the evolutionary attack.
        """
        if type(model).__name__ == "PytorchObjectDetector":
            self.model = model.model
        else:
            self.model = model
        self.obj_class = obj_class
        self.loader = loader
        self.batch_iterator = batch_iterator
        self.df = df
        if self.df == None:
            self.df = loader.dataframe
        self.device = device
        self.iou_thresh = iou_thresh
        self.nms_thresh = nms_thresh
        self.return_nms = return_nms
        self.use_preds = use_preds
        self.attacker = EvoDAttack(self.model, self.obj_class, **attkwargs)
        self.verbose = verbose
        self.atk_meta = {
            "model_name": str(model.__class__.__name__),
            "iou_thresh": str(iou_thresh),
            "nms_thresh": str(nms_thresh),
        }

        for arg in attkwargs:
            print("arg:", arg)

    def _save_bbox_ims(self, image, target_box, fname, out_path=""):
        # TODO don't fix the save location
        import torchvision.transforms as transforms
        import os

        arr = image
        im = transforms.ToPILImage()(arr)

        directory = "./perts/"

        if not os.path.exists(directory):
            os.makedirs(directory)
        im.save("perts/" + fname)

    def run(self, num_samples, save_perts=False):
        """
        #TODO add save perturbation option here.
        Run the black box robustness evaluation on a set number of samples.

        :param num_samples: number of samples containing the attack class to run attack on.
        :return: the dataframe and the ids of the attacked samples.
        """
        # TODO: Should the dataset task checker be in here?
        tracker = 0
        attacked_ids = []
        batch_enum = enumerate(self.batch_iterator)

        while tracker < num_samples:
            batch_id, (ims, labs, ids) = next(batch_enum)

            if tracker > 0 and batch_id == 0:
                break

            for idx in range(len(labs)):
                sample_id = ids[idx]
                data_dict = {}
                if self.use_preds == False:
                    perturb, gt, preds = self.attacker.attack(ims[idx], labs[idx])
                else:
                    perturb, gt, preds = self.attacker.attack(ims[idx], None)
                # make sure we have a ground truth object
                if gt != None:
                    # Add the perturbation to the image
                    pert_im = ims[idx] + perturb.to(self.device)
                    # get predictions on perturbed image
                    pert_preds = self.model([pert_im])[0]
                    # Apply non-maximal suppression to predictions
                    nms_pert_preds = nms_pred_reduce(pert_preds, self.nms_thresh)
                    nms_preds = nms_pred_reduce(preds, self.nms_thresh)
                    # check to see if gt object has disappeared
                    flip = check_flip(
                        nms_pert_preds,
                        gt,
                        self.obj_class,
                        self.iou_thresh,
                        self.nms_thresh,
                    )
                    # Get imagewise AP for class
                    troj_ap, _ = get_class_ap(
                        self.obj_class, nms_preds, labs[idx], self.iou_thresh
                    )
                    adv_troj_ap, _ = get_class_ap(
                        self.obj_class, nms_pert_preds, labs[idx], self.iou_thresh
                    )
                    # compute perturbation norm
                    pert_vec = perturb.view(-1)
                    linf_pert = torch.norm(pert_vec, p=np.inf)
                    tracker += 1
                    if self.return_nms == True:
                        (
                            orig_boxes,
                            orig_labels,
                            orig_scores,
                        ) = get_pred_data(nms_preds)
                        (
                            pert_boxes,
                            pert_labels,
                            pert_scores,
                        ) = get_pred_data(nms_pert_preds)
                    else:
                        (
                            orig_boxes,
                            orig_labels,
                            orig_scores,
                        ) = get_pred_data(preds)
                        (
                            pert_boxes,
                            pert_labels,
                            pert_scores,
                        ) = get_pred_data(pert_preds)
                    gt_boxes = gt["boxes"].detach().cpu().numpy()
                    data_dict = {
                        "flip": flip,
                        "TmAP": troj_ap.item(),
                        "Adv_TmAP": adv_troj_ap.item(),
                        "Linf": linf_pert.item(),
                        "iou_thresh": self.iou_thresh,
                        "nms_thresh": self.nms_thresh,
                        'pred_boxes':np.expand_dims(orig_boxes, axis=0).tolist(),
                        'pred_labels': np.expand_dims(orig_labels, axis=1).tolist(),
                        'pred_scores': np.expand_dims(orig_scores, axis=1).tolist(),
                        'adversarial_boxes': np.expand_dims(pert_boxes, axis=0).tolist(),
                        'adversarial_labels': np.expand_dims(pert_labels, axis=1).tolist(),
                        'adversarial_scores': np.expand_dims(pert_scores, axis=1).tolist(),
                        'gt_box':np.expand_dims(gt_boxes, axis=0).tolist()
                        }

                    self.df = log_to_dataframe(self.df, sample_id, data_dict)

                    attacked_ids.append(sample_id)
                    if save_perts:
                        self._save_bbox_ims(
                            pert_im, gt_boxes, self.df["file_name"][sample_id]
                        )
            return self.df, attacked_ids



import torch


class FGSMObjectDetection:
    def __init__(self, model, epsilon, alpha, num_iter):
        '''
        :param model: an instance of a PytorchObjectDetector object.
        :param epsilon: Maximum allowable perturbation.
        :param alpha: Step size
        :param num_iter: Number of steps
        '''
        #TODO allow user to pick certain losses from the ,odel loss dictionary to maximize.
        self.model = model
        self.epsilon = epsilon
        self.alpha = alpha
        self.num_iter = num_iter

    def generate(self, X, Y, return_examples=True, return_losses=True):
        '''
        :param X: A list of images (Pytorch tensors) to attack.
        :param Y: A list of image annotation dictionaries.
        :param return_examples: If true, returns the examples, else just returns the perturbations.
        :param return_losses: Whether or not to return the list of loss dictionaries (each dictionary contains the initial
        loss and final loss)
        :returns: A list of either adversarial examples or just the perturbations, along with the losses if return_losses=True
        '''
        adv_X = []
        loss_dicts = []
        for idx in range(len(X)):
            x = X[idx]
            y = Y[idx]
            losses = {}
            init_loss = model.compute_loss([x], [y], grad=False, reduce=True)
            losses["initial_loss"] = init_loss.item()
            delta = torch.nn.init.normal_(
                torch.zeros_like(x, requires_grad=True), mean=0, std=0.001
            )
            for t in range(num_iter):
                loss = model.compute_loss([x + delta], [y], grad=True, reduce=True)
                loss.backward()
                delta.data = (delta + x.shape[0] * alpha * delta.grad.data).clamp(
                    -epsilon, epsilon
                )
                delta.grad.zero_()
            losses["final_loss"] = loss.item()
            if return_examples:
                adv_X.append(x + delta.detach())
            else:
                adv_X.append(delta.detach())
            loss_dicts.append(losses)
        return adv_X, loss_dicts




import numpy as np
from ..metrics.mAP_utils import MakePredAnnotBoxDataframes, box_ious, ClassAP


def mAP(dataframe, iou_thresh=0.5, adversarial=False):
    # TODO documentation
    # compute mean average precision from dataframe containing bounding boxes
    ground_truth_df, predictions_df = MakePredAnnotBoxDataframes(dataframe, adversarial)
    ground_truth_df, predictions_df = box_ious(ground_truth_df, predictions_df)
    ap_dict = ClassAP(ground_truth_df, predictions_df)
    return np.mean(list(ap_dict.values()))


def true_min_pert(labels, preds, original, perturbation, losses, norm=np.inf):
    """
    Finds the smallest perturbation which flips the label. If no perturbation is found, sets the minimum to 0.5 (since
    the image will become entirely grey at such a perturbation).

    :param original_ims: Original inputs
    :param outs_array: Adversarial examples
    :param losses: Adversarial losses
    :param preds: Predictions on Adversarial Examples
    :param norm: the L^p norm to use.
    :return: minimum perturbation, the loss for the perturbation, and the prediction
    """

    tile_list = [original for i in range(perturbation.shape[1])]
    tiled = np.stack(tile_list)
    # swap the first and second axis
    tiled = np.swapaxes(tiled, 0, 1)
    if tiled.shape[1] != 1:
        assert tiled[0, 0, 0, 0, 0] == tiled[0, 1, 0, 0, 0]

    # reshape outputs and original images to collection of vectors
    flattened_shape = (
        perturbation.shape[0],
        perturbation.shape[1],
        perturbation.shape[2] * perturbation.shape[3] * perturbation.shape[4],
    )
    flattened_outs = np.reshape(perturbation, flattened_shape)
    flattened_original = np.reshape(tiled, flattened_shape)

    # subtract the original from the perturbed to get the perturbation vector
    perturbations = flattened_outs - flattened_original
    perturbation_norms = np.linalg.norm(perturbations, norm, axis=2)

    tiled_labels = [labels for i in range(preds.shape[1])]
    tiled_labels = np.stack(tiled_labels)
    # swap the first and second axis
    tiled_labels = np.swapaxes(tiled_labels, 0, 1)
    # If the label isn't flipped, set min pert to 0.5
    perturbation_norms[tiled_labels == np.argmax(preds, axis=2)] = 0.5
    min_per_sample_idx = np.argmin(perturbation_norms, axis=1)

    min_pert_losses = []
    min_pert_preds = []
    min_pert_outs = []
    for idx in range(len(min_per_sample_idx)):
        min_pert_outs.append(perturbation[idx, min_per_sample_idx[idx]])
        min_pert_losses.append(losses[idx, min_per_sample_idx[idx]])
        min_pert_preds.append(preds[idx, min_per_sample_idx[idx]])

    min_pert_outs = np.asarray(min_pert_outs)
    min_pert_preds = np.asarray(min_pert_preds)
    min_pert_losses = np.asarray(min_pert_losses)
    return min_pert_outs, min_pert_preds, min_pert_losses



import requests
from typing import Optional, List
from .client_utils import (
    assert_valid_name,
    raise_resp_exception_error,
    requests_retry,
    _upload_local_files,
    _upload_local_zip,
    validate_dataservices_args,
)
import json


class TrojClient:
    def __init__(
        self,
        *,
        api_endpoint: str = "https://wct55o2t7c.execute-api.ca-central-1.amazonaws.com/prod/api/v1",
        # api_endpoint: str = "http://localhost:8080/api/v1",
        # api_endpoint: str = "https://wuw1nar7mf.execute-api.ca-central-1.amazonaws.com/dev/api/v1",
        **kwargs,
    ) -> "Client":
        # self._creds_id_token = None
        self._creds_refresh_token = None
        self._creds_api_key = None
        self.api_endpoint = api_endpoint
        # self.cognito_client_id = "46ca3qqhhkos18k3hjqn10b4uc"
        self.cognito_client_id = "643hgaovbhf36dol8ihpfr6513"
        requests_retry.hooks["response"].append(
            self.reauth
        )  # https://github.com/psf/requests/issues/4747 - Important for Retry vs urllib3

    def _get_creds_headers(self):
        """
        Get appropriate request headers for the currently set credentials.

        Raises:
            Exception: No credentials set.
        """
        if self._creds_id_token:
            return {
                "Authorization": f"Bearer {self._creds_id_token}",
                "x-api-key": f"{self._creds_api_key}",
            }
        else:
            raise Exception("No credentials set.")

    def set_credentials(
        self,
        *,
        id_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        """
        Set credentials for the client.

        :param id_token (str, optional): Used by the client to authenticate the user.
        :param refresh_token (str, optional): Used to refresh the ID Token.
        :param api_key (str, optional): Used to gain access to API.

        Raises:
            Exception: Invalid credential combination provided.
        """

        # TODO: Change to require id_token and api_key together
        if id_token is not None:
            self._creds_id_token = id_token
        else:
            raise Exception("Please provide an ID Token.")
        if refresh_token is not None:
            self._creds_refresh_token = refresh_token
        else:
            raise Exception("Please provide a Refresh Token.")
        if api_key is not None:
            self._creds_api_key = api_key
        else:
            raise Exception("Please provide an API Key.")

    def test_api_endpoint(self):
        try:
            r = requests_retry.get(
                "https://wuw1nar7mf.execute-api.ca-central-1.amazonaws.com/dev/ping"
            )
            return r.status_code
        except Exception as exc:
            raise Exception(f"test_api_endpoint error: {exc}")

    def create_project(self, project_name: str):
        """
        Create a new project via the REST API.

        :param project_name (str): Name you want to give your project.
        """

        assert_valid_name(project_name)

        data = {"project_name": project_name}

        r = requests_retry.post(
            f"{self.api_endpoint}/projects",
            headers=self._get_creds_headers(),
            json=data,
        )

        raise_resp_exception_error(r)
        return {"status_code": r.status_code, "data": r.json()}

    def get_projects(self):
        """
        Get data about the users projects.
        """

        r = requests_retry.get(
            f"{self.api_endpoint}/projects",
            headers=self._get_creds_headers(),
        )

        raise_resp_exception_error(r)
        return {"status_code": r.status_code, "data": r.json()}

    def project_exists(self, project_name: str):
        """
        Check if a project exists.

        :param project_name (str): Name of project to check.

        Returns:
            Dict[int, dict]: dict(data) will either be False or the project itself.
        """

        r = requests_retry.get(
            f"{self.api_endpoint}/projects/{project_name}",
            headers=self._get_creds_headers(),
        )
        exists = False
        if r.status_code == 200:
            exists = True
        raise_resp_exception_error(r)
        return exists, {"status_code": r.status_code, "data": r.json()}

    # def project_exists(self, project_name: str):
    #     """
    #     Check if a project exists.

    #     :param project_name (str): Name of project to check.

    #     Returns:
    #         Dict[int, dict]: dict(data) will either be False or the project itself.
    #     """

    #     r = requests_retry.get(
    #         f"{self.api_endpoint}/projects/{project_name}",
    #         headers=self._get_creds_headers(),
    #     )

    #     raise_resp_exception_error(r)
    #     return {"status_code": r.status_code, "data": r.json()}

    def delete_project(self, project_name: str):
        """
        Try to delete a project.

        :param project_name (str): Name of the project to be deleted.
        """
        exists, res = self.project_exists(project_name)
        if not exists:
            raise Exception(f"Project '{project_name}' does not exist.")

        r = requests_retry.delete(
            f"{self.api_endpoint}/projects/{project_name}",
            headers=self._get_creds_headers(),
        )

        raise_resp_exception_error(r)
        return {"status_code": r.status_code, "data": r.json()}

    def create_dataset(self, project_name: str, dataset_name: str):
        """
        Create a new dataset under an already existing prroject.

        :param project_name (str): Name of the project you want to create the dataset in.
        :param project_name (str): Name you want to give your dataset.
        """

        # TODO: Add task choice when creating dataset
        assert_valid_name(dataset_name)
        proj_exists, project_data = self.project_exists(project_name)

        data = {
            "project_uuid": project_data["data"]["project_uuid"],
            "dataset_name": dataset_name,
        }

        r = requests_retry.post(
            f"{self.api_endpoint}/datasets",
            headers=self._get_creds_headers(),
            json=data,
        )

        raise_resp_exception_error(r)
        return {"status_code": r.status_code, "data": r.json()}
        # End of create_dataset()

    def get_project_datasets(self, project_name: str):
        """
        Get info about existing datasets for a specific project.

        :param project_name (str): Name of the project you want to find datasets under
        """

        r = requests_retry.get(
            f"{self.api_endpoint}/projects/{project_name}/datasets",
            headers=self._get_creds_headers(),
        )

        raise_resp_exception_error(r)
        return {"status_code": r.status_code, "data": r.json()}

    def dataset_exists(self, project_name: str, dataset_name: str):
        """
        Check if a dataset exists.

        Args:
            project_name (str): Name of project dataset is under.
            dataset_name (str): Name of dataset to check.

        Returns:
            Dict[int, dict]: dict(data) will either be False or the dataset itself
        """
        proj_exists, project_data = self.project_exists(project_name)

        if not proj_exists:
            raise Exception(f"Project '{project_name}' does not exist.")

        r = requests_retry.get(
            f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}",
            headers=self._get_creds_headers(),
        )

        raise_resp_exception_error(r)
        exists = False
        if r.status_code == 200:
            exists = True
            
        return exists, {"status_code": r.status_code, "data": r.json()}

    # def dataset_exists(self, project_name: str, dataset_name: str):
    #     """
    #     Check if a dataset exists.

    #     Args:
    #         project_name (str): Name of project dataset is under.
    #         dataset_name (str): Name of dataset to check.

    #     Returns:
    #         Dict[int, dict]: dict(data) will either be False or the dataset itself
    #     """
    #     proj_exists, project_data = self.new_project_exists(project_name)

    #     if not proj_exists:
    #         raise Exception(f"Project '{project_name}' does not exist.")

    #     r = requests_retry.get(
    #         f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}",
    #         headers=self._get_creds_headers(),
    #     )

    #     raise_resp_exception_error(r)
    #     return {"status_code": r.status_code, "data": r.json()}

    def delete_dataset(self, project_name: str, dataset_name: str):
        """
        Try to delete a dataset

        :param project_name (str): Name of the project dataset is under.
        :param dataset_name (str): Name of the dataset to be deleted.
        """
        dset_exists, res = self.dataset_exists(project_name, dataset_name)
        if not dset_exists:
            raise Exception(
                f"Dataset '{dataset_name}' does not exist in project '{project_name}'."
            )

        r = requests_retry.delete(
            f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}",
            headers=self._get_creds_headers(),
        )

        raise_resp_exception_error(r)
        return {"status_code": r.status_code, "data": r.json()}

    def upload_dataset_files(
        self, abs_root_path: str, project_name: str, dataset_name: str
    ):
        """
        Upload a folder of files.

        :param abs_root_path: Path to local folder of images. (files to be uploaded)
        :param project_name (str): Name of project.
        :param dataset_name (str): Name of dataset.
        """
        file_paths = self.make_files_list(abs_root_path)
        

        download_urls = []
        get_upload_path = f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}/fetch_upload_dataset_url"

        download_urls = _upload_local_files(
            file_paths,
            get_upload_path,
            self._get_creds_headers(),
            "",  # suffix
            "",  # prefix
            delete_after_upload=False,
        )

        return download_urls

    def make_files_list(self, abs_root_path):
        import os
        file_paths = []
        for root, dirs, files in os.walk(os.path.abspath(abs_root_path)):
            for file in files:
                file_paths.append(os.path.join(root, file))

        return file_paths

    def upload_dataset_files_zipped(
        self, zip_filepath: str, project_name: str, dataset_name: str
    ):
        """
        Upload a folder of zipped files.

        :param zip_filepath (str): The local folder of zipped images. (files to be uploaded)
        :param project_name (str): Name of project.
        :param dataset_name (str): Name of dataset.
        """
        download_urls = []
        get_upload_path = f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}/fetch_upload_dataset_url_zip"
        complete_upload_path = f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}/fetch_upload_dataset_url_zip/complete"
        download_urls = _upload_local_zip(
            zip_filepath,
            get_upload_path,
            complete_upload_path,
            self._get_creds_headers(),
            "",  # suffix
            "",  # prefix
            delete_after_upload=False,
        )

        return download_urls

    def upload_df_results(self, project_name: str, dataset_name: str, dataframe: dict):
        """
        Uploads dataframe results.

        :param project_name (str): Project name.
        :param dataset_name (str): Dataset name.
        :param dataframe (dict): JSONified dataframe.

        Returns:
            Dict[int, bool]: status_code and bool whether success/fail to upload
        """
        try:
            
            dset_exists, dataset = self.dataset_exists(project_name, dataset_name)

            if dataset["data"] is not False:
                r = requests_retry.post(
                    f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}/fetch_upload_dataframe_url",
                    headers=self._get_creds_headers(),
                )

                raise_resp_exception_error(r)

                s3_payload_dict = r.json()

                file_to_save = json.dumps(dataframe).encode("utf-8")
                file_name_on_s3 = s3_payload_dict["fields"]["key"]

                files = {"file": (file_name_on_s3, file_to_save)}
                upload_response = requests.post(
                    s3_payload_dict["url"], data=s3_payload_dict["fields"], files=files
                )

                print(f"Upload response: {upload_response.status_code}")

                return {"status_code": upload_response.status_code}
            else:
                raise Exception(
                    "Something went wrong. Double check the project and dataset names."
                )
        except Exception as exc:
            raise Exception(f"post_dataframe error: {exc}")

    def refresh_tokens(self):

        url = "https://troj.auth.ca-central-1.amazoncognito.com/oauth2/token"

        payload = f"grant_type=refresh_token&client_id={self.cognito_client_id}&refresh_token={self._creds_refresh_token}"
        requests.utils.quote(payload)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9",
        }

        response = requests.request("POST", url, headers=headers, data=payload)
        print("res:", response.json())
        self._creds_id_token = json.loads(response.text)["id_token"]

        return response

    def reauth(self, res, *args, **kwargs):
        """Hook to re-authenticate whenever authentication expires."""
        if res.status_code == requests.codes.forbidden:
            if res.request.headers.get("REATTEMPT"):
                res.raise_for_status()
            self.refresh_tokens()
            req = res.request
            req.headers["REATTEMPT"] = 1
            req = self.auth_inside_hook(req)
            res = requests_retry.send(req)
            return res

    def auth_inside_hook(self, req):
        """Set the authentication token for the premade request during reauth attempts inside the retry hook."""
        req.headers["Authorization"] = f"Bearer {self._creds_id_token}"
        return req

    def train_model(
        self,
        project_name: str,
        dataset_name: str,
        architecture_name: Optional[str] = "resnet18",
        pretrained: Optional[bool] = True,
        num_classes: Optional[int] = 1000,
        global_pool: Optional[str] = None,
        finetune: Optional[bool] = True,
        resume_from_checkpoint: Optional[bool] = False,
        load_previous_model: Optional[str] = None,
        optimizer: Optional[str] = "Adam",
        learning_rate: Optional[float] = 0.001,
        momentum: Optional[float] = 0.9,
        dampening: Optional[float] = 0,
        weight_decay: Optional[float] = 5e-4,
        nesterov: Optional[bool] = False,
        betas: Optional[tuple] = (0.9, 0.999),
        adam_eps: Optional[float] = 1e-08,
        amsgrad: Optional[bool] = False,
        lr_scheduler: Optional[str] = "cosine",
        T_0: Optional[int] = 10,
        T_mult: Optional[int] = 1,
        eta_min: Optional[float] = 0,
        factor: Optional[float] = 0.1,
        patience: Optional[int] = 10,
        threshold: Optional[float] = 0.0001,
        cooldown: Optional[int] = 0,
        min_lr: Optional[float] = 0,
        normalization_stats: Optional[str] = None,
        batch_size: Optional[int] = 64,
        model_name: Optional[str] = None,
        early_stop_patience: Optional[int] = 7,
        num_epochs: Optional[int] = 50,
        log_interval: Optional[int] = 100,
        use_workers: Optional[bool] = False,
        log_wandb: Optional[bool] = False,
        wandb_key: Optional[str] = "",
    ):
        """
        Transfer learns a model from users' uploaded dataset.

        :param project_name (str): Project name
        :param dataset_name (str): Dataset name
        :param architecture_name, default='resnet18', help="Name of the input model architecture", type=str)
        :param pretrained, default=True, help='Whether or not to use pretrained weights for model', type=bool)
        :param num_classes, default=1000, type=int, help='Number of classes for the model')
        :param global_pool, default=None, metavar='POOL', help='Global pool type, one of (fast, avg, max, avgmax, avgmaxc). Model default if None.')
        :param finetune, default=True, type=bool, help='whether or not to only train a new output layer')
        :param resume_from_checkpoint, default=False, help='If true, the loaded model file in the laod previous model is expected to containing an optimizer and epoch as well as the state dict.')
        :param load_previous_model, default=None, help='If to load a model state dict (not from a checkpoint), set to model weights path.')
        :param optimizer, default='Adam', type=str, help='Optimizer to use')
        :param learning_rate, default=0.001, type=float, help='Initial Learning Rate')
        :param momentum, default=0.9, type=float, help='Momentum for SGD')
        :param dampening, default=0, type=float, help='Dampening for SGD')
        :param weight_decay, default=5e-4, type=float, help='Weight Decay for SGD')
        :param nesterov, default=False, type=bool, help='Whether or not to use Nesterov momentum for SGD')
        :param betas, default=(0.9, 0.999), type=tuple, help='Beta values for Adam/AdamW')
        :param adam_eps, default=1e-08, type=float, help='epsilon values for Adam/AdamW')
        :param amsgrad, default=False, type=bool, help='Whether or not to use AMSgrad when using Adam/AdamW')
        :param lr_scheduler, default='cosine', help='Learning rate scheduler. Options are None, plateu, cosine, and one_cycle')
        :param T_0, default=10, type=int, help='Number of iterations for original cosine annealing schedule')
        :param T_mult, default=1, type=int, help='')
        :param eta_min, default=0, type=float,)
        :param factor, default=0.1, type=float, help='Plateu decay factor')
        :param patience, default=10, type=int, help='Plateu decay patience')
        :param threshold, default=0.0001, type=float, help=' Threshold for measuring the new optimum')
        :param cooldown, default=0, type=int, help='Number of epochs to wait before resuming normal operation after lr has been reduced.')
        :param min_lr, default=0, type=float, help='Minimum LR allowed from plateu decay')
        :param normalization_stats, default=None, help='If model is not pretrained, stats to use for normalization. Either None, a pickle of two lists, or -1. If -1, we compute them. If using pretrained model, we use the default normalization.')
        :param batch_size, default=64, type=int, help='Size of training batches')
        :param model_name, default=None, help='The name to save the model as. If None, name is autogenerated')
        :param early_stop_patience, default=7, type=int, help='How many epochs before stopping when loss no longer decreases')
        :param num_epochs, default=50, type=int, help='Number of training epochs')
        :param log_interval, default=100, type=int, help='Number of batches between data logs')
        :param use_workers, default=False, type=bool, help='Whether or not to use multiple workers. On Windows, must be false.')
        :param log_wandb, default=False, type=bool, help='Logging for testing purposes')
        :param wandb_key, default='nokey', help='wandb key for logging.')

        Returns:
            Dict[int, Dict]: Status code and response of batch submit_job.
        """

        dset_exists, dataset = self.dataset_exists(project_name, dataset_name)

        # Argument validation
        validate_dataservices_args(
            num_classes=num_classes,
            learning_rate=learning_rate,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            betas=betas,
            adam_eps=adam_eps,
            T_0=T_0,
            T_mult=T_mult,
            eta_min=eta_min,
            factor=factor,
            patience=patience,
            threshold=threshold,
            cooldown=cooldown,
            min_lr=min_lr,
            batch_size=batch_size,
            early_stop_patience=early_stop_patience,
            num_epochs=num_epochs,
            log_interval=log_interval,
        )

        if dataset["data"] is not False:
            data = {
                "project_name": project_name,
                "dataset_name": dataset_name,
                "architecture_name": architecture_name,
                "pretrained": pretrained,
                "num_classes": num_classes,
                "global_pool": global_pool,
                "finetune": finetune,
                "resume_from_checkpoint": resume_from_checkpoint,
                "load_previous_model": load_previous_model,
                "optimizer": optimizer,
                "learning_rate": learning_rate,
                "momentum": momentum,
                "dampening": dampening,
                "weight_decay": weight_decay,
                "nesterov": nesterov,
                "betas": betas,
                "adam_eps": adam_eps,
                "amsgrad": amsgrad,
                "lr_scheduler": lr_scheduler,
                "t_0": T_0,
                "t_mult": T_mult,
                "eta_min": eta_min,
                "factor": factor,
                "patience": patience,
                "threshold": threshold,
                "cooldown": cooldown,
                "min_lr": min_lr,
                "normalization_stats": normalization_stats,
                "batch_size": batch_size,
                "model_name": model_name,
                "early_stop_patience": early_stop_patience,
                "num_epochs": num_epochs,
                "log_interval": log_interval,
                "use_workers": use_workers,
                "log_wandb": log_wandb,
                "wandb_key": wandb_key,
            }

            r = requests_retry.post(
                f"{self.api_endpoint}/projects/{project_name}/datasets/{dataset_name}/train",
                headers=self._get_creds_headers(),
                params=data,
            )

            raise_resp_exception_error(r)
            return {"status_code": r.status_code, "data": r.json()}
        else:
            raise Exception(
                "Something went wrong. Double check the project and dataset names."
            )



import numpy as np


def get_pred_data(preds):
    """

    :param preds: predictions from object detection model.
    :return: the boxes, labels, and scores as tensors.
    """
    boxes = preds["boxes"].detach().cpu().numpy()
    labels = preds["labels"].detach().cpu().numpy()
    scores = preds["scores"].detach().cpu().numpy()
    return boxes, labels, scores


def compute_Lp_distance(x1, x2, p=np.inf):
    """
    compute Lp distance of a collection of images against another collection.

    :param x1: image collection 1
    :param x2: image collection 2
    :param p: p norm
    :return: tensor of distances
    """
    x1 = np.reshape(x1, (x1.shape[0], -1))
    x2 = np.reshape(x2, (x2.shape[0], -1))
    difference_vect = x1 - x2
    lp_distance = np.linalg.norm(difference_vect, p, axis=1)
    return lp_distance



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="troj",
    version="0.2.0.6.0",
    packages=find_packages(),
    author="TrojAI",
    author_email="stan.petley@troj.ai",
    description="TrojAI provides the troj Python convenience package to allow users to integrate TrojAI adversarial protections and robustness metrics seamlessly into their AI development pipelines.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://troj.ai",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "torch",
        "tensorflow",
        "requests",
        "urllib3",
        "adversarial-robustness-toolbox==1.7.0",
        "opencv-python",
        "torchvision",
    ],
    python_requires=">=3.6",
    extras_require={
        "dev": [
            "pytest>=6.2",
        ],
    },

)



"""
Utilities for attacks in the trojai package.

"""


import torchvision.ops
import torch


def nms_reduction(predictions, thresh=0.05):
    """
    Performs non-maximal suppression on output dictionary of network.

    :param predictions: A dictionary containing the keys [boxes, labels, scores] where boxes are the bounding boxes for
    predicted objects in the form [x,y, x+w, y+h] where w,h are the width and height of the bounding box. Labels are the
    predicted labels, and scores are the confidence of prediction.
    :param thresh: Threshold with which to perform the suppression.
    :return: Supressed predictions.
    """
    # returns predictions after NMS
    fixed_indices = torchvision.ops.nms(
        predictions["boxes"], predictions["scores"], thresh
    )
    reduced_predictions = {
        "boxes": predictions["boxes"][fixed_indices],
        "labels": predictions["labels"][fixed_indices],
        "scores": predictions["scores"][fixed_indices],
    }
    return reduced_predictions


def get_objects_of_type(predictions, object_class, inc_scores=True):
    """
    Gets predictions for objects only of a given type.

    :param predictions: Original predictions. Model typically outputs list, this expects only elements of that list.
    :param object_class: Object class id
    :param inc_scores: whether or not to include the scores in the output
    :return: The prediction dictionary containing only the desired class.
    """
    # extracts predictions for the class we wish to evaluate
    labels = predictions["labels"]
    obj_ids = (labels == object_class).nonzero()
    if inc_scores == True:
        reduced_predictions = {
            "boxes": predictions["boxes"][obj_ids],
            "labels": predictions["labels"][obj_ids],
            "scores": predictions["scores"][obj_ids],
        }
    else:
        reduced_predictions = {
            "boxes": predictions["boxes"][obj_ids],
            "labels": predictions["labels"][obj_ids],
        }
    return reduced_predictions


def detect_dict_from_idx(inp_dict, ids):
    """
    Return subdictionary of prediction dictionary based on sets of value ids.

    :param inp_dict: Input dictionary
    :param ids: list of ids
    :return: Dictionary containing the values at the locations specified by the ids.
    """
    key_list = list(inp_dict.keys())
    new_dict = {}
    for key in key_list:
        new_dict[key] = inp_dict[key][ids]
    return new_dict


def get_gt_indices(indicator_matrix):
    """
    Gets the ground truth objects corresponding to the predictions
    :param indicator_matrix: Two dimensional matrix of booleans
    :return: ground truth objects tensor
    """
    ground_truth_objs = []
    for idx in range(indicator_matrix.shape[0]):
        gt = indicator_matrix[idx].nonzero()
        if gt.shape[0] == 0:
            gt_obj = -1
        else:
            gt_obj = gt[0]
        ground_truth_objs.append(gt_obj)
    return torch.tensor(ground_truth_objs)


def get_tp_fp(gt_inds):
    """
    Used for the troj imagewise Average Precision score. gets the true positives and false positives as a list. If a
     ground truth object is predicted more than once, every prediction after the first is a false positive.
    :param gt_inds: Ground truth index tensor
    :return: A tensor where a zero entry represents a false positive and a 1 represents a true positive.
    """
    # gets the true positives and false positives as a list. If a ground truth object is predicted more than once, every prediction after the first is
    # a false positive.
    taken_idxs = []
    tp_fp_list = []
    for idx in gt_inds:
        if idx not in taken_idxs and idx != -1:
            taken_idxs.append(idx)
            tp_fp_list.append(1)
        else:
            tp_fp_list.append(0)
    return torch.tensor(tp_fp_list)


def get_fn_indicators(indicator_matrix):
    """
    Gets number of false negatives.
    :param indicator_matrix:  Two dimensional matrix of booleans
    :return: a tensor where the entry is 1 if that index is a false negative and 0 otherwise.
    """
    # check for false negatives
    fn_inds = []
    indices = torch.tensor([i for i in range(indicator_matrix.shape[1])])
    for idx in indices:
        # 1 if false negative, 0 otherwise
        val = 1
        for obj_indicators in indicator_matrix:
            if obj_indicators[idx] == True:
                val = 0
                break
        fn_inds.append(val)
    return torch.tensor(fn_inds)


def accumulated_tp_fp(tp_fp_indicators):
    """
    :param tp_fp_indicators: the index indicator tensor for the true/false positives
    :return: two tensors, one containing the true positive indices and one containing the false positive indices.
    """
    tp_list = []
    fp_list = []
    for idx in range(tp_fp_indicators.shape[0]):
        tp = torch.sum(tp_fp_indicators[: idx + 1])
        tp_list.append(tp)
        fp_list.append(tp_fp_indicators.shape[0] - tp)
    return torch.tensor(tp_list), torch.tensor(fp_list)


def AP(precisions, recalls):
    # AP for PASCAL VOC 2010
    vals = []
    for i in range(recalls.shape[0]):
        p_interp = torch.max(precisions[i:])
        vals.append(p_interp)
    return torch.mean(torch.tensor(vals))


def get_class_ap(class_id, predicts, annots, iou_thresh=0.5):
    # for troj AP
    # Gets average precision of a class on a single image at a certain threshold.
    # if predicts is list of dictionaries, take first entry
    if isinstance(predicts, list):
        predicts = predicts[0]
    # get objects of the given type from annotations and predictions
    obj_preds = get_objects_of_type(predicts, class_id, inc_scores=True)
    obj_annots = get_objects_of_type(annots, class_id, inc_scores=False)
    # compute IOU values between objects of class class_id in annotations and all predictions made by model
    iou_matrix = torchvision.ops.box_iou(
        obj_annots["boxes"].squeeze(1), obj_preds["boxes"].squeeze(1)
    )
    # Get the indices where the IOU of the bounding box is greater than the threshold.
    positive_det = iou_matrix >= iou_thresh
    # store the (transpose) positive detection matrix in the obj_preds dictionary
    obj_preds["iou_thresh_indicator"] = torch.transpose(positive_det, 0, 1)
    # now, sort obj_preds dictionary based on the descending confidence scores
    sorting_args = torch.argsort(obj_preds["scores"].squeeze(1), descending=True)
    sorted_obj_preds = {}
    for key in obj_preds.keys():
        sorted_obj_preds[key] = obj_preds[key][sorting_args]
    # Get the indices for ground truth objects corresponding to predictions. A -1 means a false positive.
    ind_matrix = sorted_obj_preds["iou_thresh_indicator"]
    false_negs = get_fn_indicators(ind_matrix)
    obj_annots["false_neg"] = false_negs
    gto = get_gt_indices(ind_matrix)
    sorted_obj_preds["gt_idx"] = gto
    # indicate whether prediction is true positive or false positive. In OD, true negative has no meaning, so it is ignored.
    tp_fp_tensor = get_tp_fp(sorted_obj_preds["gt_idx"])
    sorted_obj_preds["tp_fp_indicator"] = tp_fp_tensor
    # get accumulated true and false positives
    acc_tp, acc_fp = accumulated_tp_fp(tp_fp_tensor)
    sorted_obj_preds["acc_tp"] = acc_tp
    sorted_obj_preds["acc_fp"] = acc_fp
    # compute precision and recall at different confidences
    recalls = acc_tp / ind_matrix.shape[1]
    precisions = acc_tp / tp_fp_tensor.shape[0]
    sorted_obj_preds["recall"] = recalls
    sorted_obj_preds["precision"] = precisions
    av_prec = AP(precisions, recalls)
    return av_prec, sorted_obj_preds


def check_flip(predictions, ground_truth, obj_cat, iou_thresh=0.5, nms_thresh=0.05):
    """
    Checks for flips of the gt object.

    :param predictions: Model predictions (after nms)
    :param ground_truth: ground truth object.
    :param obj_cat: Deprecated
    :param iou_thresh: Minimum iou thresh to count as detection.
    :param nms_thresh: Deprecated
    :return:
    """

    nms_reduced_preds = predictions  # detect_dict_from_idx(predictions, pred_ids)
    reduced_preds = get_objects_of_type(nms_reduced_preds, obj_cat)
    if reduced_preds["boxes"].nelement() == 0:
        flip = 1
        return flip
        # return flip, nms_reduced_preds
    ious = torchvision.ops.box_iou(
        ground_truth["boxes"], reduced_preds["boxes"].squeeze(dim=1)
    )
    flip = 1
    if torch.max(ious) >= iou_thresh:
        flip = 0
    return flip


def nms_pred_reduce(predictions, nms_thresh=0.05):
    """
    Performs nms on predictions.
    :param predictions: original predictions.
    :param nms_thresh: Suppression threshold
    :return: Suppressed predictions
    """
    pred_ids = torchvision.ops.nms(
        predictions["boxes"], predictions["scores"], nms_thresh
    )
    nms_reduced_preds = detect_dict_from_idx(predictions, pred_ids)
    return nms_reduced_preds



from collections import defaultdict


class Logger():
    def __init__(self, *args):
        self.logger_dict = defaultdict(list)
    
    """
    Return the logger 

    :return : the logger dict  
    """
    def get_logger(self):
        return self.logger_dict

    """
    Log a dictionary

    :param in_dict: a dictionary to be added to the logger
    """
    def log(self, in_dict):
        for key in in_dict.keys():
            self.logger_dict[key].append(in_dict[key])
        